from pydantic import BaseModel, Field


class Restriction(BaseModel):
    code: str = Field(..., alias="restrictionCode")
    description: str = Field(..., alias="restrictionDescription")


__all__ = ('Restriction',)



import aiohttp
import pytest

from glQiwiApi import QiwiWrapper, InvalidData
from glQiwiApi.core import RequestManager
from glQiwiApi.core.web_hooks.dispatcher import Dispatcher
from glQiwiApi.qiwi.settings import QiwiRouter, QiwiKassaRouter

try:
    from asynctest import CoroutineMock, patch
except ImportError:
    from unittest.mock import AsyncMock as CoroutineMock, patch  # type: ignore

pytestmark = pytest.mark.asyncio


class TestAiohttpSession:

    async def test_create_api(self):
        from tests.types.dataset import API_DATA
        api = QiwiWrapper(**API_DATA)

        # initially session is None,
        # it's done  to save performance and
        # it creates new session when you make a request (API method call)
        assert api.session is None

        # such a session is created under the hood(when you call API method)
        api._requests.create_session()

        # And now it's aiohttp.ClientSession instance
        assert isinstance(api.session, aiohttp.ClientSession)

        assert isinstance(api._router, QiwiRouter)
        assert isinstance(api._requests, RequestManager)
        assert isinstance(api._p2p_router, QiwiKassaRouter)
        assert isinstance(api.dispatcher, Dispatcher)

        # close session =)
        await api.close()

    async def test_create_api_with_wrong_data(self):
        from tests.types.dataset import WRONG_API_DATA
        with pytest.raises(InvalidData):
            QiwiWrapper(**WRONG_API_DATA)

    async def test_close_session(self):
        from tests.types.dataset import API_DATA
        api = QiwiWrapper(**API_DATA)

        api._requests.create_session()

        aiohttp_session = api.session

        with patch("aiohttp.ClientSession.close",
                   new=CoroutineMock()) as mocked_close:
            await aiohttp_session.close()
            mocked_close.assert_called_once()

        await api.close()



import datetime
import uuid
from typing import Dict

import aiohttp
import pytest

from glQiwiApi import QiwiWrapper, sync
from glQiwiApi import types


@pytest.fixture(name="api")
def sync_api_fixture(credentials: Dict[str, str]):
    _wrapper = QiwiWrapper(**credentials, without_context=True)
    yield _wrapper


def test_sync_get_balance(api: QiwiWrapper):
    result = sync(api.get_balance)
    assert isinstance(result, types.Sum)


@pytest.mark.parametrize("params", [
    {"amount": 1},
    {"amount": 1, "comment": "test_comment"},
    {
        "amount": 1,
        "comment": "test_comment",
        "life_time": datetime.datetime.now() + datetime.timedelta(hours=5)
    },
    {
        "amount": 1,
        "comment": "test_comment",
        "life_time": datetime.datetime.now() + datetime.timedelta(hours=5),
        "bill_id": str(uuid.uuid4())
    }
])
def test_sync_create_p2p_bill(api: QiwiWrapper, params: dict):
    result = sync(api.create_p2p_bill, **params)
    assert isinstance(result, types.Bill)


class SyncApiSessionTest:

    @pytest.mark.last
    def test_is_session_closing(self, api: QiwiWrapper):
        # Send request to API
        sync(api.get_balance)

        api_session = api.session

        assert isinstance(api_session, aiohttp.ClientSession)

        # Send new request to API

        sync(api.get_balance)

        new_session = api.session

        assert api_session != new_session



import inspect
import operator
from typing import Any, Callable

from glQiwiApi import types
from .config import CF, E


def special_comparator(func1: CF, func2: CF, event: Any) -> Any:
    """ custom a & b (and) operator """
    validated = func1(event)
    if validated is not True:
        return False
    return operator.and_(validated, func2(event))


def xor(func1: CF, func2: CF, event: Any) -> Any:
    """ custom a ^ b (xor) operator """
    funcs = (func1, func2)
    for func in funcs:
        if func(event):
            return True
    return False


def or_(func1: CF, func2: CF, event: Any) -> Any:
    """ Custom a | b (or) operator """
    return func1(event) or func2(event)


class Filter:
    """
    Base Filter object, callback container

    """

    __name__: str

    def __init__(self, function: CF):
        self.__name__ = f"Filter around <{function!r}>"

        self.function: Callable = function
        self.awaitable: bool = inspect.iscoroutinefunction(
            function
        ) or inspect.isawaitable(function)

    def __eq__(self, other: Any) -> "Filter":
        return _sing_filter(self, lambda filter1: operator.eq(filter1, other))

    def __ne__(self, other: Any) -> "Filter":
        return _sing_filter(self, lambda filter1: operator.ne(filter1, other))

    def __invert__(self) -> "Filter":
        return _sing_filter(self, operator.not_)

    def __xor__(self, other: "Filter") -> "Filter":
        return _compose_filter(self, other, xor)

    def __and__(self, other: "Filter") -> "Filter":
        return _compose_filter(self, other, special_comparator)

    def __or__(self, other: "Filter") -> "Filter":
        return _compose_filter(self, other, or_)


def _compose_filter(
        filter1: Filter,
        filter2: Filter,
        operator_: Any
) -> Filter:
    """
    Compose two filters

    :param filter1:
    :param filter2:
    :param operator_:
    """
    if (not isinstance(filter1, Filter)) | (not isinstance(filter2, Filter)):
        raise ValueError(
            f"Cannot compare non-Filter object with Filter, "
            f"got filter1={filter1.__name__} "
            f"operator={operator_.__name__} "
            f"filter2={filter2.__name__}"
        )

    func: CF

    if filter1.awaitable & filter2.awaitable:

        async def func(event: E) -> Any:
            return operator_(
                await filter1.function(event), await filter2.function(event)
            )

    elif filter1.awaitable ^ filter2.awaitable:
        is_f1_async = filter1.awaitable
        async_func = filter1 if is_f1_async else filter2
        sync_func = filter2 if is_f1_async else filter1

        async def func(event: E) -> Any:
            return operator_(
                await async_func.function(event), sync_func.function(event)
            )

    else:

        def func(event: E) -> Any:
            return operator_(filter1.function, filter2.function, event)

    return Filter(func)


def _sing_filter(filter1: Filter, operator_) -> Filter:
    func: CF

    if filter1.awaitable:

        async def func(event: E) -> Any:
            return operator_(await filter1.function(event))

    else:

        def func(event: E) -> Any:
            return operator_(filter1.function(event))

    return Filter(func)


# Default filter for transaction handler
transaction_webhook_filter = Filter(
    lambda update: isinstance(update, (types.WebHook, types.Transaction))
)

# Default filter for bill handler
bill_webhook_filter = Filter(
    lambda update: isinstance(update, types.Notification)
)

__all__ = (
    "Filter",
    "transaction_webhook_filter",
    "bill_webhook_filter"
)



from typing import Union, Optional

from pydantic import BaseModel


class CurrencyModel(BaseModel):
    code: str
    decimal_digits: int
    name: str
    name_plural: str
    rounding: Union[int, float]
    symbol: str
    symbol_native: str
    isoformat: Optional[str] = None


codes_number = {
    "008": "ALL",
    "012": "DZD",
    "032": "ARS",
    "036": "AUD",
    "044": "BSD",
    "048": "BHD",
    "050": "BDT",
    "051": "AMD",
    "052": "BBD",
    "060": "BMD",
    "064": "BTN",
    "068": "BOB",
    "072": "BWP",
    "084": "BZD",
    "090": "SBD",
    "096": "BND",
    "104": "MMK",
    "108": "BIF",
    "116": "KHR",
    "124": "CAD",
    "132": "CVE",
    "136": "KYD",
    "144": "LKR",
    "152": "CLP",
    "156": "CNY",
    "170": "COP",
    "174": "KMF",
    "188": "CRC",
    "191": "HRK",
    "192": "CUP",
    "203": "CZK",
    "208": "DKK",
    "214": "DOP",
    "222": "SVC",
    "230": "ETB",
    "232": "ERN",
    "238": "FKP",
    "242": "FJD",
    "262": "DJF",
    "270": "GMD",
    "292": "GIP",
    "320": "GTQ",
    "324": "GNF",
    "328": "GYD",
    "332": "HTG",
    "340": "HNL",
    "344": "HKD",
    "348": "HUF",
    "352": "ISK",
    "356": "INR",
    "360": "IDR",
    "364": "IRR",
    "368": "IQD",
    "376": "ILS",
    "388": "JMD",
    "392": "JPY",
    "398": "KZT",
    "400": "JOD",
    "404": "KES",
    "408": "KPW",
    "410": "KRW",
    "414": "KWD",
    "417": "KGS",
    "418": "LAK",
    "422": "LBP",
    "426": "LSL",
    "430": "LRD",
    "434": "LYD",
    "446": "MOP",
    "454": "MWK",
    "458": "MYR",
    "462": "MVR",
    "480": "MUR",
    "484": "MXN",
    "496": "MNT",
    "498": "MDL",
    "504": "MAD",
    "512": "OMR",
    "516": "NAD",
    "524": "NPR",
    "532": "ANG",
    "533": "AWG",
    "548": "VUV",
    "554": "NZD",
    "558": "NIO",
    "566": "NGN",
    "578": "NOK",
    "586": "PKR",
    "590": "PAB",
    "598": "PGK",
    "600": "PYG",
    "604": "PEN",
    "608": "PHP",
    "634": "QAR",
    "643": "RUB",
    "646": "RWF",
    "654": "SHP",
    "682": "SAR",
    "690": "SCR",
    "694": "SLL",
    "702": "SGD",
    "704": "VND",
    "706": "SOS",
    "710": "ZAR",
    "728": "SSP",
    "748": "SZL",
    "752": "SEK",
    "756": "CHF",
    "760": "SYP",
    "764": "THB",
    "776": "TOP",
    "780": "TTD",
    "784": "AED",
    "788": "TND",
    "800": "UGX",
    "807": "MKD",
    "818": "EGP",
    "826": "GBP",
    "834": "TZS",
    "840": "USD",
    "858": "UYU",
    "860": "UZS",
    "882": "WST",
    "886": "YER",
    "901": "TWD",
    "929": "MRU",
    "930": "STN",
    "931": "CUC",
    "932": "ZWL",
    "934": "TMT",
    "936": "GHS",
    "937": "VEF",
    "938": "SDG",
    "940": "UYI",
    "941": "RSD",
    "943": "MZN",
    "944": "AZN",
    "946": "RON",
    "947": "CHE",
    "948": "CHW",
    "949": "TRY",
    "950": "XAF",
    "951": "XCD",
    "952": "XOF",
    "953": "XPF",
    "960": "XDR",
    "965": "XUA",
    "967": "ZMW",
    "968": "SRD",
    "969": "MGA",
    "970": "COU",
    "971": "AFN",
    "972": "TJS",
    "973": "AOA",
    "974": "BYR",
    "975": "BGN",
    "976": "CDF",
    "977": "BAM",
    "978": "EUR",
    "979": "MXV",
    "980": "UAH",
    "981": "GEL",
    "984": "BOV",
    "985": "PLN",
    "986": "BRL",
    "990": "CLF",
    "994": "XSU",
    "997": "USN",
}
described = {
    "AED": CurrencyModel(
        code="AED",
        decimal_digits=2,
        name="United Arab Emirates Dirham",
        name_plural="UAE dirhams",
        rounding=0,
        symbol="AED",
        symbol_native="د.إ.\u200f",
        isoformat="784",
    ),
    "AFN": CurrencyModel(
        code="AFN",
        decimal_digits=0,
        name="Afghan Afghani",
        name_plural="Afghan Afghanis",
        rounding=0,
        symbol="Af",
        symbol_native="؋",
        isoformat="971",
    ),
    "ALL": CurrencyModel(
        code="ALL",
        decimal_digits=0,
        name="Albanian Lek",
        name_plural="Albanian lekë",
        rounding=0,
        symbol="ALL",
        symbol_native="Lek",
        isoformat="008",
    ),
    "AMD": CurrencyModel(
        code="AMD",
        decimal_digits=0,
        name="Armenian Dram",
        name_plural="Armenian drams",
        rounding=0,
        symbol="AMD",
        symbol_native="դր.",
        isoformat="051",
    ),
    "ARS": CurrencyModel(
        code="ARS",
        decimal_digits=2,
        name="Argentine Peso",
        name_plural="Argentine pesos",
        rounding=0,
        symbol="AR$",
        symbol_native="$",
        isoformat="032",
    ),
    "AUD": CurrencyModel(
        code="AUD",
        decimal_digits=2,
        name="Australian Dollar",
        name_plural="Australian dollars",
        rounding=0,
        symbol="AU$",
        symbol_native="$",
        isoformat="036",
    ),
    "AZN": CurrencyModel(
        code="AZN",
        decimal_digits=2,
        name="Azerbaijani Manat",
        name_plural="Azerbaijani manats",
        rounding=0,
        symbol="man.",
        symbol_native="ман.",
        isoformat="944",
    ),
    "BAM": CurrencyModel(
        code="BAM",
        decimal_digits=2,
        name="Bosnia-Herzegovina Convertible Mark",
        name_plural="Bosnia-Herzegovina convertible marks",
        rounding=0,
        symbol="KM",
        symbol_native="KM",
        isoformat="977",
    ),
    "BDT": CurrencyModel(
        code="BDT",
        decimal_digits=2,
        name="Bangladeshi Taka",
        name_plural="Bangladeshi takas",
        rounding=0,
        symbol="Tk",
        symbol_native="৳",
        isoformat="050",
    ),
    "BGN": CurrencyModel(
        code="BGN",
        decimal_digits=2,
        name="Bulgarian Lev",
        name_plural="Bulgarian leva",
        rounding=0,
        symbol="BGN",
        symbol_native="лв.",
        isoformat="975",
    ),
    "BHD": CurrencyModel(
        code="BHD",
        decimal_digits=3,
        name="Bahraini Dinar",
        name_plural="Bahraini dinars",
        rounding=0,
        symbol="BD",
        symbol_native="د.ب.\u200f",
        isoformat="048",
    ),
    "BIF": CurrencyModel(
        code="BIF",
        decimal_digits=0,
        name="Burundian Franc",
        name_plural="Burundian francs",
        rounding=0,
        symbol="FBu",
        symbol_native="FBu",
        isoformat="108",
    ),
    "BND": CurrencyModel(
        code="BND",
        decimal_digits=2,
        name="Brunei Dollar",
        name_plural="Brunei dollars",
        rounding=0,
        symbol="BN$",
        symbol_native="$",
        isoformat="096",
    ),
    "BOB": CurrencyModel(
        code="BOB",
        decimal_digits=2,
        name="Bolivian Boliviano",
        name_plural="Bolivian bolivianos",
        rounding=0,
        symbol="Bs",
        symbol_native="Bs",
        isoformat="068",
    ),
    "BRL": CurrencyModel(
        code="BRL",
        decimal_digits=2,
        name="Brazilian Real",
        name_plural="Brazilian reals",
        rounding=0,
        symbol="R$",
        symbol_native="R$",
        isoformat="986",
    ),
    "BWP": CurrencyModel(
        code="BWP",
        decimal_digits=2,
        name="Botswanan Pula",
        name_plural="Botswanan pulas",
        rounding=0,
        symbol="BWP",
        symbol_native="P",
        isoformat="072",
    ),
    "BYR": CurrencyModel(
        code="BYR",
        decimal_digits=0,
        name="Belarusian Ruble",
        name_plural="Belarusian rubles",
        rounding=0,
        symbol="BYR",
        symbol_native="BYR",
        isoformat="974",
    ),
    "BZD": CurrencyModel(
        code="BZD",
        decimal_digits=2,
        name="Belize Dollar",
        name_plural="Belize dollars",
        rounding=0,
        symbol="BZ$",
        symbol_native="$",
        isoformat="084",
    ),
    "CAD": CurrencyModel(
        code="CAD",
        decimal_digits=2,
        name="Canadian Dollar",
        name_plural="Canadian dollars",
        rounding=0,
        symbol="CA$",
        symbol_native="$",
        isoformat="124",
    ),
    "CDF": CurrencyModel(
        code="CDF",
        decimal_digits=2,
        name="Congolese Franc",
        name_plural="Congolese francs",
        rounding=0,
        symbol="CDF",
        symbol_native="FrCD",
        isoformat="976",
    ),
    "CHF": CurrencyModel(
        code="CHF",
        decimal_digits=2,
        name="Swiss Franc",
        name_plural="Swiss francs",
        rounding=0.05,
        symbol="CHF",
        symbol_native="CHF",
        isoformat="756",
    ),
    "CLP": CurrencyModel(
        code="CLP",
        decimal_digits=0,
        name="Chilean Peso",
        name_plural="Chilean pesos",
        rounding=0,
        symbol="CL$",
        symbol_native="$",
        isoformat="152",
    ),
    "CNY": CurrencyModel(
        code="CNY",
        decimal_digits=2,
        name="Chinese Yuan",
        name_plural="Chinese yuan",
        rounding=0,
        symbol="CN¥",
        symbol_native="CN¥",
        isoformat="156",
    ),
    "COP": CurrencyModel(
        code="COP",
        decimal_digits=0,
        name="Colombian Peso",
        name_plural="Colombian pesos",
        rounding=0,
        symbol="CO$",
        symbol_native="$",
        isoformat="170",
    ),
    "CRC": CurrencyModel(
        code="CRC",
        decimal_digits=0,
        name="Costa Rican Colón",
        name_plural="Costa Rican colóns",
        rounding=0,
        symbol="₡",
        symbol_native="₡",
        isoformat="188",
    ),
    "CVE": CurrencyModel(
        code="CVE",
        decimal_digits=2,
        name="Cape Verdean Escudo",
        name_plural="Cape Verdean escudos",
        rounding=0,
        symbol="CV$",
        symbol_native="CV$",
        isoformat="132",
    ),
    "CZK": CurrencyModel(
        code="CZK",
        decimal_digits=2,
        name="Czech Republic Koruna",
        name_plural="Czech Republic korunas",
        rounding=0,
        symbol="Kč",
        symbol_native="Kč",
        isoformat="203",
    ),
    "DJF": CurrencyModel(
        code="DJF",
        decimal_digits=0,
        name="Djiboutian Franc",
        name_plural="Djiboutian francs",
        rounding=0,
        symbol="Fdj",
        symbol_native="Fdj",
        isoformat="262",
    ),
    "DKK": CurrencyModel(
        code="DKK",
        decimal_digits=2,
        name="Danish Krone",
        name_plural="Danish kroner",
        rounding=0,
        symbol="Dkr",
        symbol_native="kr",
        isoformat="208",
    ),
    "DOP": CurrencyModel(
        code="DOP",
        decimal_digits=2,
        name="Dominican Peso",
        name_plural="Dominican pesos",
        rounding=0,
        symbol="RD$",
        symbol_native="RD$",
        isoformat="214",
    ),
    "DZD": CurrencyModel(
        code="DZD",
        decimal_digits=2,
        name="Algerian Dinar",
        name_plural="Algerian dinars",
        rounding=0,
        symbol="DA",
        symbol_native="د.ج.\u200f",
        isoformat="012",
    ),
    "EEK": CurrencyModel(
        code="EEK",
        decimal_digits=2,
        name="Estonian Kroon",
        name_plural="Estonian kroons",
        rounding=0,
        symbol="Ekr",
        symbol_native="kr",
        isoformat=None,
    ),
    "EGP": CurrencyModel(
        code="EGP",
        decimal_digits=2,
        name="Egyptian Pound",
        name_plural="Egyptian pounds",
        rounding=0,
        symbol="EGP",
        symbol_native="ج.م.\u200f",
        isoformat="818",
    ),
    "ERN": CurrencyModel(
        code="ERN",
        decimal_digits=2,
        name="Eritrean Nakfa",
        name_plural="Eritrean nakfas",
        rounding=0,
        symbol="Nfk",
        symbol_native="Nfk",
        isoformat="232",
    ),
    "ETB": CurrencyModel(
        code="ETB",
        decimal_digits=2,
        name="Ethiopian Birr",
        name_plural="Ethiopian birrs",
        rounding=0,
        symbol="Br",
        symbol_native="Br",
        isoformat="230",
    ),
    "EUR": CurrencyModel(
        code="EUR",
        decimal_digits=2,
        name="Euro",
        name_plural="euros",
        rounding=0,
        symbol="€",
        symbol_native="€",
        isoformat="978",
    ),
    "GBP": CurrencyModel(
        code="GBP",
        decimal_digits=2,
        name="British Pound Sterling",
        name_plural="British pounds sterling",
        rounding=0,
        symbol="£",
        symbol_native="£",
        isoformat="826",
    ),
    "GEL": CurrencyModel(
        code="GEL",
        decimal_digits=2,
        name="Georgian Lari",
        name_plural="Georgian laris",
        rounding=0,
        symbol="GEL",
        symbol_native="GEL",
        isoformat="981",
    ),
    "GHS": CurrencyModel(
        code="GHS",
        decimal_digits=2,
        name="Ghanaian Cedi",
        name_plural="Ghanaian cedis",
        rounding=0,
        symbol="GH₵",
        symbol_native="GH₵",
        isoformat="936",
    ),
    "GNF": CurrencyModel(
        code="GNF",
        decimal_digits=0,
        name="Guinean Franc",
        name_plural="Guinean francs",
        rounding=0,
        symbol="FG",
        symbol_native="FG",
        isoformat="324",
    ),
    "GTQ": CurrencyModel(
        code="GTQ",
        decimal_digits=2,
        name="Guatemalan Quetzal",
        name_plural="Guatemalan quetzals",
        rounding=0,
        symbol="GTQ",
        symbol_native="Q",
        isoformat="320",
    ),
    "HKD": CurrencyModel(
        code="HKD",
        decimal_digits=2,
        name="Hong Kong Dollar",
        name_plural="Hong Kong dollars",
        rounding=0,
        symbol="HK$",
        symbol_native="$",
        isoformat="344",
    ),
    "HNL": CurrencyModel(
        code="HNL",
        decimal_digits=2,
        name="Honduran Lempira",
        name_plural="Honduran lempiras",
        rounding=0,
        symbol="HNL",
        symbol_native="L",
        isoformat="340",
    ),
    "HRK": CurrencyModel(
        code="HRK",
        decimal_digits=2,
        name="Croatian Kuna",
        name_plural="Croatian kunas",
        rounding=0,
        symbol="kn",
        symbol_native="kn",
        isoformat="191",
    ),
    "HUF": CurrencyModel(
        code="HUF",
        decimal_digits=0,
        name="Hungarian Forint",
        name_plural="Hungarian forints",
        rounding=0,
        symbol="Ft",
        symbol_native="Ft",
        isoformat="348",
    ),
    "IDR": CurrencyModel(
        code="IDR",
        decimal_digits=0,
        name="Indonesian Rupiah",
        name_plural="Indonesian rupiahs",
        rounding=0,
        symbol="Rp",
        symbol_native="Rp",
        isoformat="360",
    ),
    "ILS": CurrencyModel(
        code="ILS",
        decimal_digits=2,
        name="Israeli New Sheqel",
        name_plural="Israeli new sheqels",
        rounding=0,
        symbol="₪",
        symbol_native="₪",
        isoformat="376",
    ),
    "INR": CurrencyModel(
        code="INR",
        decimal_digits=2,
        name="Indian Rupee",
        name_plural="Indian rupees",
        rounding=0,
        symbol="Rs",
        symbol_native="টকা",
        isoformat="356",
    ),
    "IQD": CurrencyModel(
        code="IQD",
        decimal_digits=0,
        name="Iraqi Dinar",
        name_plural="Iraqi dinars",
        rounding=0,
        symbol="IQD",
        symbol_native="د.ع.\u200f",
        isoformat="368",
    ),
    "IRR": CurrencyModel(
        code="IRR",
        decimal_digits=0,
        name="Iranian Rial",
        name_plural="Iranian rials",
        rounding=0,
        symbol="IRR",
        symbol_native="﷼",
        isoformat="364",
    ),
    "ISK": CurrencyModel(
        code="ISK",
        decimal_digits=0,
        name="Icelandic Króna",
        name_plural="Icelandic krónur",
        rounding=0,
        symbol="Ikr",
        symbol_native="kr",
        isoformat="352",
    ),
    "JMD": CurrencyModel(
        code="JMD",
        decimal_digits=2,
        name="Jamaican Dollar",
        name_plural="Jamaican dollars",
        rounding=0,
        symbol="J$",
        symbol_native="$",
        isoformat="388",
    ),
    "JOD": CurrencyModel(
        code="JOD",
        decimal_digits=3,
        name="Jordanian Dinar",
        name_plural="Jordanian dinars",
        rounding=0,
        symbol="JD",
        symbol_native="د.أ.\u200f",
        isoformat="400",
    ),
    "JPY": CurrencyModel(
        code="JPY",
        decimal_digits=0,
        name="Japanese Yen",
        name_plural="Japanese yen",
        rounding=0,
        symbol="¥",
        symbol_native="￥",
        isoformat="392",
    ),
    "KES": CurrencyModel(
        code="KES",
        decimal_digits=2,
        name="Kenyan Shilling",
        name_plural="Kenyan shillings",
        rounding=0,
        symbol="Ksh",
        symbol_native="Ksh",
        isoformat="404",
    ),
    "KHR": CurrencyModel(
        code="KHR",
        decimal_digits=2,
        name="Cambodian Riel",
        name_plural="Cambodian riels",
        rounding=0,
        symbol="KHR",
        symbol_native="៛",
        isoformat="116",
    ),
    "KMF": CurrencyModel(
        code="KMF",
        decimal_digits=0,
        name="Comorian Franc",
        name_plural="Comorian francs",
        rounding=0,
        symbol="CF",
        symbol_native="FC",
        isoformat="174",
    ),
    "KRW": CurrencyModel(
        code="KRW",
        decimal_digits=0,
        name="South Korean Won",
        name_plural="South Korean won",
        rounding=0,
        symbol="₩",
        symbol_native="₩",
        isoformat="410",
    ),
    "KWD": CurrencyModel(
        code="KWD",
        decimal_digits=3,
        name="Kuwaiti Dinar",
        name_plural="Kuwaiti dinars",
        rounding=0,
        symbol="KD",
        symbol_native="د.ك.\u200f",
        isoformat="414",
    ),
    "KZT": CurrencyModel(
        code="KZT",
        decimal_digits=2,
        name="Kazakhstani Tenge",
        name_plural="Kazakhstani tenges",
        rounding=0,
        symbol="KZT",
        symbol_native="тңг.",
        isoformat="398",
    ),
    "LBP": CurrencyModel(
        code="LBP",
        decimal_digits=0,
        name="Lebanese Pound",
        name_plural="Lebanese pounds",
        rounding=0,
        symbol="LB£",
        symbol_native="ل.ل.\u200f",
        isoformat="422",
    ),
    "LKR": CurrencyModel(
        code="LKR",
        decimal_digits=2,
        name="Sri Lankan Rupee",
        name_plural="Sri Lankan rupees",
        rounding=0,
        symbol="SLRs",
        symbol_native="SL Re",
        isoformat="144",
    ),
    "LTL": CurrencyModel(
        code="LTL",
        decimal_digits=2,
        name="Lithuanian Litas",
        name_plural="Lithuanian litai",
        rounding=0,
        symbol="Lt",
        symbol_native="Lt",
        isoformat=None,
    ),
    "LVL": CurrencyModel(
        code="LVL",
        decimal_digits=2,
        name="Latvian Lats",
        name_plural="Latvian lati",
        rounding=0,
        symbol="Ls",
        symbol_native="Ls",
        isoformat=None,
    ),
    "LYD": CurrencyModel(
        code="LYD",
        decimal_digits=3,
        name="Libyan Dinar",
        name_plural="Libyan dinars",
        rounding=0,
        symbol="LD",
        symbol_native="د.ل.\u200f",
        isoformat="434",
    ),
    "MAD": CurrencyModel(
        code="MAD",
        decimal_digits=2,
        name="Moroccan Dirham",
        name_plural="Moroccan dirhams",
        rounding=0,
        symbol="MAD",
        symbol_native="د.م.\u200f",
        isoformat="504",
    ),
    "MDL": CurrencyModel(
        code="MDL",
        decimal_digits=2,
        name="Moldovan Leu",
        name_plural="Moldovan lei",
        rounding=0,
        symbol="MDL",
        symbol_native="MDL",
        isoformat="498",
    ),
    "MGA": CurrencyModel(
        code="MGA",
        decimal_digits=0,
        name="Malagasy Ariary",
        name_plural="Malagasy Ariaries",
        rounding=0,
        symbol="MGA",
        symbol_native="MGA",
        isoformat="969",
    ),
    "MKD": CurrencyModel(
        code="MKD",
        decimal_digits=2,
        name="Macedonian Denar",
        name_plural="Macedonian denari",
        rounding=0,
        symbol="MKD",
        symbol_native="MKD",
        isoformat="807",
    ),
    "MMK": CurrencyModel(
        code="MMK",
        decimal_digits=0,
        name="Myanma Kyat",
        name_plural="Myanma kyats",
        rounding=0,
        symbol="MMK",
        symbol_native="K",
        isoformat="104",
    ),
    "MOP": CurrencyModel(
        code="MOP",
        decimal_digits=2,
        name="Macanese Pataca",
        name_plural="Macanese patacas",
        rounding=0,
        symbol="MOP$",
        symbol_native="MOP$",
        isoformat="446",
    ),
    "MUR": CurrencyModel(
        code="MUR",
        decimal_digits=0,
        name="Mauritian Rupee",
        name_plural="Mauritian rupees",
        rounding=0,
        symbol="MURs",
        symbol_native="MURs",
        isoformat="480",
    ),
    "MXN": CurrencyModel(
        code="MXN",
        decimal_digits=2,
        name="Mexican Peso",
        name_plural="Mexican pesos",
        rounding=0,
        symbol="MX$",
        symbol_native="$",
        isoformat="484",
    ),
    "MYR": CurrencyModel(
        code="MYR",
        decimal_digits=2,
        name="Malaysian Ringgit",
        name_plural="Malaysian ringgits",
        rounding=0,
        symbol="RM",
        symbol_native="RM",
        isoformat="458",
    ),
    "MZN": CurrencyModel(
        code="MZN",
        decimal_digits=2,
        name="Mozambican Metical",
        name_plural="Mozambican meticals",
        rounding=0,
        symbol="MTn",
        symbol_native="MTn",
        isoformat="943",
    ),
    "NAD": CurrencyModel(
        code="NAD",
        decimal_digits=2,
        name="Namibian Dollar",
        name_plural="Namibian dollars",
        rounding=0,
        symbol="N$",
        symbol_native="N$",
        isoformat="516",
    ),
    "NGN": CurrencyModel(
        code="NGN",
        decimal_digits=2,
        name="Nigerian Naira",
        name_plural="Nigerian nairas",
        rounding=0,
        symbol="₦",
        symbol_native="₦",
        isoformat="566",
    ),
    "NIO": CurrencyModel(
        code="NIO",
        decimal_digits=2,
        name="Nicaraguan Córdoba",
        name_plural="Nicaraguan córdobas",
        rounding=0,
        symbol="C$",
        symbol_native="C$",
        isoformat="558",
    ),
    "NOK": CurrencyModel(
        code="NOK",
        decimal_digits=2,
        name="Norwegian Krone",
        name_plural="Norwegian kroner",
        rounding=0,
        symbol="Nkr",
        symbol_native="kr",
        isoformat="578",
    ),
    "NPR": CurrencyModel(
        code="NPR",
        decimal_digits=2,
        name="Nepalese Rupee",
        name_plural="Nepalese rupees",
        rounding=0,
        symbol="NPRs",
        symbol_native="नेरू",
        isoformat="524",
    ),
    "NZD": CurrencyModel(
        code="NZD",
        decimal_digits=2,
        name="New Zealand Dollar",
        name_plural="New Zealand dollars",
        rounding=0,
        symbol="NZ$",
        symbol_native="$",
        isoformat="554",
    ),
    "OMR": CurrencyModel(
        code="OMR",
        decimal_digits=3,
        name="Omani Rial",
        name_plural="Omani rials",
        rounding=0,
        symbol="OMR",
        symbol_native="ر.ع.\u200f",
        isoformat="512",
    ),
    "PAB": CurrencyModel(
        code="PAB",
        decimal_digits=2,
        name="Panamanian Balboa",
        name_plural="Panamanian balboas",
        rounding=0,
        symbol="B/.",
        symbol_native="B/.",
        isoformat="590",
    ),
    "PEN": CurrencyModel(
        code="PEN",
        decimal_digits=2,
        name="Peruvian Nuevo Sol",
        name_plural="Peruvian nuevos soles",
        rounding=0,
        symbol="S/.",
        symbol_native="S/.",
        isoformat="604",
    ),
    "PHP": CurrencyModel(
        code="PHP",
        decimal_digits=2,
        name="Philippine Peso",
        name_plural="Philippine pesos",
        rounding=0,
        symbol="₱",
        symbol_native="₱",
        isoformat="608",
    ),
    "PKR": CurrencyModel(
        code="PKR",
        decimal_digits=0,
        name="Pakistani Rupee",
        name_plural="Pakistani rupees",
        rounding=0,
        symbol="PKRs",
        symbol_native="₨",
        isoformat="586",
    ),
    "PLN": CurrencyModel(
        code="PLN",
        decimal_digits=2,
        name="Polish Zloty",
        name_plural="Polish zlotys",
        rounding=0,
        symbol="zł",
        symbol_native="zł",
        isoformat="985",
    ),
    "PYG": CurrencyModel(
        code="PYG",
        decimal_digits=0,
        name="Paraguayan Guarani",
        name_plural="Paraguayan guaranis",
        rounding=0,
        symbol="₲",
        symbol_native="₲",
        isoformat="600",
    ),
    "QAR": CurrencyModel(
        code="QAR",
        decimal_digits=2,
        name="Qatari Rial",
        name_plural="Qatari rials",
        rounding=0,
        symbol="QR",
        symbol_native="ر.ق.\u200f",
        isoformat="634",
    ),
    "RON": CurrencyModel(
        code="RON",
        decimal_digits=2,
        name="Romanian Leu",
        name_plural="Romanian lei",
        rounding=0,
        symbol="RON",
        symbol_native="RON",
        isoformat="946",
    ),
    "RSD": CurrencyModel(
        code="RSD",
        decimal_digits=0,
        name="Serbian Dinar",
        name_plural="Serbian dinars",
        rounding=0,
        symbol="din.",
        symbol_native="дин.",
        isoformat="941",
    ),
    "RUB": CurrencyModel(
        code="RUB",
        decimal_digits=2,
        name="Russian Ruble",
        name_plural="Russian rubles",
        rounding=0,
        symbol="RUB",
        symbol_native="руб.",
        isoformat="643",
    ),
    "RWF": CurrencyModel(
        code="RWF",
        decimal_digits=0,
        name="Rwandan Franc",
        name_plural="Rwandan francs",
        rounding=0,
        symbol="RWF",
        symbol_native="FR",
        isoformat="646",
    ),
    "SAR": CurrencyModel(
        code="SAR",
        decimal_digits=2,
        name="Saudi Riyal",
        name_plural="Saudi riyals",
        rounding=0,
        symbol="SR",
        symbol_native="ر.س.\u200f",
        isoformat="682",
    ),
    "SDG": CurrencyModel(
        code="SDG",
        decimal_digits=2,
        name="Sudanese Pound",
        name_plural="Sudanese pounds",
        rounding=0,
        symbol="SDG",
        symbol_native="SDG",
        isoformat="938",
    ),
    "SEK": CurrencyModel(
        code="SEK",
        decimal_digits=2,
        name="Swedish Krona",
        name_plural="Swedish kronor",
        rounding=0,
        symbol="Skr",
        symbol_native="kr",
        isoformat="752",
    ),
    "SGD": CurrencyModel(
        code="SGD",
        decimal_digits=2,
        name="Singapore Dollar",
        name_plural="Singapore dollars",
        rounding=0,
        symbol="S$",
        symbol_native="$",
        isoformat="702",
    ),
    "SOS": CurrencyModel(
        code="SOS",
        decimal_digits=0,
        name="Somali Shilling",
        name_plural="Somali shillings",
        rounding=0,
        symbol="Ssh",
        symbol_native="Ssh",
        isoformat="706",
    ),
    "SYP": CurrencyModel(
        code="SYP",
        decimal_digits=0,
        name="Syrian Pound",
        name_plural="Syrian pounds",
        rounding=0,
        symbol="SY£",
        symbol_native="ل.س.\u200f",
        isoformat="760",
    ),
    "THB": CurrencyModel(
        code="THB",
        decimal_digits=2,
        name="Thai Baht",
        name_plural="Thai baht",
        rounding=0,
        symbol="฿",
        symbol_native="฿",
        isoformat="764",
    ),
    "TND": CurrencyModel(
        code="TND",
        decimal_digits=3,
        name="Tunisian Dinar",
        name_plural="Tunisian dinars",
        rounding=0,
        symbol="DT",
        symbol_native="د.ت.\u200f",
        isoformat="788",
    ),
    "TOP": CurrencyModel(
        code="TOP",
        decimal_digits=2,
        name="Tongan Paʻanga",
        name_plural="Tongan paʻanga",
        rounding=0,
        symbol="T$",
        symbol_native="T$",
        isoformat="776",
    ),
    "TRY": CurrencyModel(
        code="TRY",
        decimal_digits=2,
        name="Turkish Lira",
        name_plural="Turkish Lira",
        rounding=0,
        symbol="TL",
        symbol_native="TL",
        isoformat="949",
    ),
    "TTD": CurrencyModel(
        code="TTD",
        decimal_digits=2,
        name="Trinidad and Tobago Dollar",
        name_plural="Trinidad and Tobago dollars",
        rounding=0,
        symbol="TT$",
        symbol_native="$",
        isoformat="780",
    ),
    "TWD": CurrencyModel(
        code="TWD",
        decimal_digits=2,
        name="New Taiwan Dollar",
        name_plural="New Taiwan dollars",
        rounding=0,
        symbol="NT$",
        symbol_native="NT$",
        isoformat="901",
    ),
    "TZS": CurrencyModel(
        code="TZS",
        decimal_digits=0,
        name="Tanzanian Shilling",
        name_plural="Tanzanian shillings",
        rounding=0,
        symbol="TSh",
        symbol_native="TSh",
        isoformat="834",
    ),
    "UAH": CurrencyModel(
        code="UAH",
        decimal_digits=2,
        name="Ukrainian Hryvnia",
        name_plural="Ukrainian hryvnias",
        rounding=0,
        symbol="₴",
        symbol_native="₴",
        isoformat="980",
    ),
    "UGX": CurrencyModel(
        code="UGX",
        decimal_digits=0,
        name="Ugandan Shilling",
        name_plural="Ugandan shillings",
        rounding=0,
        symbol="USh",
        symbol_native="USh",
        isoformat="800",
    ),
    "USD": CurrencyModel(
        code="USD",
        decimal_digits=2,
        name="US Dollar",
        name_plural="US dollars",
        rounding=0,
        symbol="$",
        symbol_native="$",
        isoformat="840",
    ),
    "UYU": CurrencyModel(
        code="UYU",
        decimal_digits=2,
        name="Uruguayan Peso",
        name_plural="Uruguayan pesos",
        rounding=0,
        symbol="$U",
        symbol_native="$",
        isoformat="858",
    ),
    "UZS": CurrencyModel(
        code="UZS",
        decimal_digits=0,
        name="Uzbekistan Som",
        name_plural="Uzbekistan som",
        rounding=0,
        symbol="UZS",
        symbol_native="UZS",
        isoformat="860",
    ),
    "VEF": CurrencyModel(
        code="VEF",
        decimal_digits=2,
        name="Venezuelan Bolívar",
        name_plural="Venezuelan bolívars",
        rounding=0,
        symbol="Bs.F.",
        symbol_native="Bs.F.",
        isoformat="937",
    ),
    "VND": CurrencyModel(
        code="VND",
        decimal_digits=0,
        name="Vietnamese Dong",
        name_plural="Vietnamese dong",
        rounding=0,
        symbol="₫",
        symbol_native="₫",
        isoformat="704",
    ),
    "XAF": CurrencyModel(
        code="XAF",
        decimal_digits=0,
        name="CFA Franc BEAC",
        name_plural="CFA francs BEAC",
        rounding=0,
        symbol="FCFA",
        symbol_native="FCFA",
        isoformat="950",
    ),
    "XOF": CurrencyModel(
        code="XOF",
        decimal_digits=0,
        name="CFA Franc BCEAO",
        name_plural="CFA francs BCEAO",
        rounding=0,
        symbol="CFA",
        symbol_native="CFA",
        isoformat="952",
    ),
    "YER": CurrencyModel(
        code="YER",
        decimal_digits=0,
        name="Yemeni Rial",
        name_plural="Yemeni rials",
        rounding=0,
        symbol="YR",
        symbol_native="ر.ي.\u200f",
        isoformat="886",
    ),
    "ZAR": CurrencyModel(
        code="ZAR",
        decimal_digits=2,
        name="South African Rand",
        name_plural="South African rand",
        rounding=0,
        symbol="R",
        symbol_native="R",
        isoformat="710",
    ),
    "ZMK": CurrencyModel(
        code="ZMK",
        decimal_digits=0,
        name="Zambian Kwacha",
        name_plural="Zambian kwachas",
        rounding=0,
        symbol="ZK",
        symbol_native="ZK",
        isoformat=None,
    ),
}



import ipaddress
import typing
from dataclasses import dataclass

DEFAULT_QIWI_WEBHOOK_PATH = "/web_hooks/qiwi/"
DEFAULT_QIWI_ROUTER_NAME = "QIWI"

DEFAULT_QIWI_BILLS_WEBHOOK_PATH = "/webhooks/qiwi/bills/"
DEFAULT_QIWI_BILLS_ROUTER_NAME = "QIWI_BILLS"

RESPONSE_TIMEOUT = 55

allowed_ips = {
    ipaddress.IPv4Network("79.142.16.0/20"),
    ipaddress.IPv4Network("195.189.100.0/22"),
    ipaddress.IPv4Network("91.232.230.0/23"),
    ipaddress.IPv4Network("91.213.51.0/24"),
}


@dataclass(frozen=True)
class Path:
    bill_path: typing.Optional[str] = None
    transaction_path: typing.Optional[str] = None


E = typing.TypeVar("E")
EventFilter = typing.Callable[[E], bool]
EventHandlerFunctor = typing.Callable[[E], typing.Awaitable[typing.Any]]
CF = typing.Callable[[E], typing.Union[typing.Awaitable[bool], bool]]

__all__ = (
    'DEFAULT_QIWI_WEBHOOK_PATH',
    'DEFAULT_QIWI_ROUTER_NAME',
    'DEFAULT_QIWI_BILLS_WEBHOOK_PATH',
    'DEFAULT_QIWI_BILLS_ROUTER_NAME',
    'RESPONSE_TIMEOUT',
    'allowed_ips',
    'Path',
    'E',
    'EventFilter',
    'EventHandlerFunctor',
    'CF'
)



from copy import deepcopy
from typing import Optional, Any, MutableMapping

from glQiwiApi.core.abstracts import AbstractRouter
from glQiwiApi.types import PaymentInfo, OrderDetails
from glQiwiApi.utils import basics as api_helper


class QiwiMasterMixin:
    """Mixin, which implements QIWI master API logic"""

    def __init__(self, router: AbstractRouter, request_manager: Any):
        self._router = router
        self._requests = request_manager

    def _auth_token(
            self,
            headers: MutableMapping,
            p2p: bool = False
    ) -> MutableMapping:
        ...

    @property
    def stripped_number(self) -> str:
        return ""

    async def buy_qiwi_master(self) -> PaymentInfo:
        """
        Метод для покупки пакета QIWI Мастер

        Для вызова методов API вам потребуется токен API QIWI Wallet
        с разрешениями на следующие действия:

        1. Управление виртуальными картами,
        2. Запрос информации о профиле кошелька,
        3. Просмотр истории платежей,
        4. Проведение платежей без SMS.

        Эти права вы можете выбрать при создании нового апи токена,
        чтобы пользоваться апи QIWI Master
        """
        url = self._router.build_url("BUY_QIWI_MASTER")
        payload = api_helper.qiwi_master_data(self.stripped_number,
                                              self._router.config.QIWI_MASTER)
        async for response in self._requests.fast().fetch(
                url=url,
                json=payload,
                method='POST',
                headers=self._auth_token(deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                ))
        ):
            return PaymentInfo.parse_obj(response.response_data)

    async def __pre_qiwi_master_request(
            self,
            card_alias: str = 'qvc-cpa'
    ) -> OrderDetails:
        """
        Метод для выпуска виртуальной карты QIWI Мастер

        :param card_alias: Тип карты
        :return: OrderDetails
        """
        url = self._router.build_url(
            "PRE_QIWI_REQUEST",
            stripped_number=self.stripped_number
        )
        async for response in self._requests.fast().fetch(
                url=url,
                headers=self._auth_token(deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                )),
                json={"cardAlias": card_alias},
                method='POST'
        ):
            return OrderDetails.parse_obj(response.response_data)

    async def _confirm_qiwi_master_request(
            self,
            card_alias: str = 'qvc-cpa'
    ) -> OrderDetails:
        """
        Подтверждение заказа выпуска карты

        :param card_alias: Тип карты
        :return: OrderDetails
        """
        details = await self.__pre_qiwi_master_request(card_alias)
        url = self._router.build_url(
            "_CONFIRM_QIWI_MASTER",
            stripped_number=self.stripped_number,
            order_id=details.order_id
        )
        async for response in self._requests.fast().fetch(
                url=url,
                headers=self._auth_token(deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                )),
                method='PUT'
        ):
            return OrderDetails.parse_obj(response.response_data)

    async def __buy_new_qiwi_card(
            self,
            **kwargs
    ) -> Optional[OrderDetails]:
        """
        Покупка карты, если она платная

        :param kwargs:
        :return: OrderDetails
        """
        kwargs.update(data=self._router.config.QIWI_MASTER)
        url, payload = api_helper.new_card_data(**kwargs)
        async for response in self._requests.fast().fetch(
                url=url,
                json=payload,
                headers=self._auth_token(deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                ))
        ):
            return OrderDetails.parse_obj(response.response_data)

    async def issue_qiwi_master_card(
            self,
            card_alias: str = 'qvc-cpa'
    ) -> Optional[OrderDetails]:
        """
        Выпуск новой карты, используя Qiwi Master API

        При выпуске карты производиться 3, а возможно 3 запроса,
        а именно по такой схеме:
            - __pre_qiwi_master_request - данный метод создает заявку
            - _confirm_qiwi_master_request - подтверждает выпуск карты
            - __buy_new_qiwi_card - покупает новую карту,
              если такая карта не бесплатна


        Подробная документация:

        https://developer.qiwi.com/ru/qiwi-wallet-personal/#qiwi-master-issue-card

        :param card_alias: Тип карты
        :return: OrderDetails
        """
        pre_response = await self._confirm_qiwi_master_request(card_alias)
        if pre_response.status == 'COMPLETED':
            return pre_response
        return await self.__buy_new_qiwi_card(
            ph_number=self.stripped_number,
            order_id=pre_response.order_id
        )

    async def _cards_qiwi_master(self):
        """
        Метод для получение списка всех ваших карт QIWI Мастер

        """
        url = self._router.build_url("CARDS_QIWI_MASTER")
        async for response in self._requests.fast().fetch(
                url=url,
                headers=self._auth_token(deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                )),
                method='GET'
        ):
            return response



from .account import Account
from .account_info import QiwiAccountInfo
from .balance import Balance
from .bill import Bill, BillError, RefundBill, Notification, P2PKeys
from .identification import Identification
from .limit import Limit
from .other import CrossRate, FreePaymentDetailsFields, PaymentMethod
from .partner import Partner
from .payment_info import PaymentInfo
from .polygon import Polygon
from .qiwi_master import OrderDetails, Card
from .stats import Statistic
from .terminal import Terminal
from .transaction import Transaction
from .webhooks import WebHookConfig, WebHook
from .restriction import Restriction

__all__ = (
    'QiwiAccountInfo',
    'Transaction',
    'Bill',
    'BillError',
    'Statistic',
    'Limit',
    'Account',
    'Balance',
    'Identification',
    'PaymentInfo',
    'OrderDetails',
    'RefundBill',
    'Polygon',
    'Terminal',
    'Partner',
    'WebHook',
    'WebHookConfig',
    'Notification',
    'P2PKeys',
    'CrossRate',
    'FreePaymentDetailsFields',
    'PaymentMethod',
    'Card',
    'Restriction'
)



import time
from dataclasses import dataclass
from typing import Union, Any, Optional

from pydantic import BaseModel, Field, validator, Extra
from glQiwiApi.utils.basics import custom_load

DEFAULT_CACHE_TIME = 0


class Sum(BaseModel):
    """
    Сумма платежа

    """
    amount: Union[int, float, str]
    currency: Any

    class Config:
        """ Pydantic config """
        json_loads = custom_load
        extra = Extra.allow

        def __str__(self) -> str:
            return '<Config class of pydantic>'

        def __repr__(self) -> str:
            return self.__str__()

    @validator("currency", pre=True, check_fields=True)
    def humanize_pay_currency(cls, v):
        from glQiwiApi.utils.currency_util import Currency
        if not isinstance(v, int):
            return v
        return Currency.get(str(v))


class OptionalSum(BaseModel):
    """ object: OptionalSum """
    value: Union[int, float]
    currency: str


class Commission(BaseModel):
    """
    Комиссия за платеж

    """
    provider_id: int = Field(alias="providerId")
    withdraw_sum: Sum = Field(alias="withdrawSum")
    enrollment_sum: Sum = Field(alias="enrollmentSum")
    qiwi_commission: Sum = Field("qwCommission")
    withdraw_to_enrollment_rate: int = Field(alias="withdrawToEnrollmentRate")

    class Config:
        """ Pydantic config """
        json_loads = custom_load

        def __str__(self) -> str:
            return '<Config class of pydantic>'

        def __repr__(self) -> str:
            return self.__str__()


class Type(BaseModel):
    """
    Базовая модель типа данных

    """
    id: str
    title: str


@dataclass
class Attributes:
    """
    Аттрибуты кэшированного запроса

    """
    headers: Optional[dict] = None
    json: Optional[dict] = None
    params: Optional[dict] = None
    data: Optional[dict] = None

    @classmethod
    def format(cls, kwargs: dict, args: tuple):
        return cls(
            **{k: kwargs.get(k) for k in args if
               isinstance(kwargs.get(k), dict)}
        )


@dataclass
class Cached:
    """
    Объект кэшированного запроса

    """
    kwargs: Attributes
    response_data: Any
    key: Optional[str]
    cache_to: Optional[Any]
    method: Optional[Any]
    cached_in: float = time.monotonic()
    status_code: Union[str, int, None] = None


__all__ = [
    'Sum',
    'OptionalSum',
    'Commission',
    'Type',
    'Cached',
    'DEFAULT_CACHE_TIME',
    'Attributes'
]



"""Main model: Partner"""
from typing import List, Optional

from pydantic import BaseModel


class Partner(BaseModel):
    """ Base partner class """
    title: str
    id: int

    maps: Optional[List[str]] = None

__all__ = ("Partner",)



import uuid
from copy import deepcopy
from typing import Any, Optional, Union, List, MutableMapping

from glQiwiApi.types import Commission, CrossRate, Sum, PaymentMethod, \
    FreePaymentDetailsFields, PaymentInfo
from glQiwiApi.utils.exceptions import InvalidCardNumber
from glQiwiApi.core.abstracts import AbstractRouter
from glQiwiApi.utils import basics as api_helper


class QiwiPaymentsMixin:
    """Provides payment QIWI API"""

    def __init__(self, requests_manager: Any, router: AbstractRouter):
        self._requests = requests_manager
        self._router: AbstractRouter = router

    def _auth_token(
            self,
            headers: MutableMapping,
            p2p: bool = False
    ) -> MutableMapping:
        ...

    async def to_wallet(
            self,
            to_number: str,
            trans_sum: Union[str, float, int],
            currency: str = '643',
            comment: str = '+comment+') -> Optional[str]:
        """
        Метод для перевода денег на другой кошелек\n
        Подробная документация:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#p2p

        :param to_number: номер получателя
        :param trans_sum: кол-во денег, которое вы хотите перевести
        :param currency: особенный код валюты
        :param comment: комментарий к платежу
        """
        data = api_helper.set_data_to_wallet(
            data=deepcopy(self._router.config.QIWI_TO_WALLET),
            to_number=to_number,
            trans_sum=trans_sum,
            currency=currency,
            comment=comment
        )
        data.headers = self._auth_token(
            headers=data.headers
        )
        async for response in self._requests.fast().fetch(
                url=self._router.build_url("TO_WALLET"),
                json=data.json,
                headers=data.headers
        ):
            return response.response_data['transaction']['id']

    async def to_card(
            self,
            trans_sum: Union[float, int],
            to_card: str
    ) -> Optional[str]:
        """
        Метод для отправки средств на карту.
        Более подробная документация:
        https://developer.qiwi.com/ru/qiwi-wallet-personal/#cards

        :param trans_sum: сумма перевода
        :param to_card: номер карты получателя
        :return:
        """
        data = api_helper.parse_card_data(
            default_data=self._router.config.QIWI_TO_CARD,
            to_card=to_card,
            trans_sum=trans_sum,
            auth_maker=self._auth_token
        )
        privat_card_id = await self._detect_card_number(card_number=to_card)
        async for response in self._requests.fast().fetch(
                url=self._router.build_url(
                    "TO_CARD", privat_card_id=privat_card_id
                ),
                headers=data.headers,
                json=data.json,
                get_json=True
        ):
            return response.response_data.get('id')

    async def _detect_card_number(self, card_number: str) -> str:
        """
        Метод для получения идентификатора карты

        https://developer.qiwi.com/ru/qiwi-wallet-personal/?python#cards
        """
        headers = deepcopy(self._router.config.DEFAULT_QIWI_HEADERS)
        headers.update(
            {
                'Content-Type': 'application/x-www-form-urlencoded'
            }
        )
        async for response in self._requests.fetch(
                url='https://qiwi.com/card/detect.action',
                headers=headers,
                method='POST',
                data={
                    'cardNumber': card_number
                }
        ):
            try:
                return response.response_data.get('message')
            except KeyError:
                raise InvalidCardNumber(
                    'Invalid card number or qiwi api is not response'
                ) from None

    async def commission(
            self,
            to_account: str,
            pay_sum: Union[int, float]
    ) -> Commission:
        """
        Возвращается полная комиссия QIWI Кошелька
        за платеж в пользу указанного провайдера
        с учетом всех тарифов по заданному набору платежных реквизитов.

        :param to_account: номер карты или киви кошелька
        :param pay_sum: сумма, за которую вы хотите узнать комиссию
        :return: Commission object
        """
        payload, special_code = api_helper.parse_commission_request_payload(
            default_data=self._router.config.COMMISSION_DATA,
            auth_maker=self._auth_token,
            to_account=to_account,
            pay_sum=pay_sum
        )
        if not isinstance(special_code, str):
            special_code = await self._detect_card_number(
                card_number=to_account
            )
        url = self._router.build_url("COMMISSION", special_code=special_code)
        async for response in self._requests.fast().fetch(
                url=url,
                headers=payload.headers,
                json=payload.json
        ):
            return Commission.parse_obj(response.response_data)

    async def get_cross_rates(self) -> List[CrossRate]:
        """
        Метод возвращает текущие курсы и кросс-курсы валют КИВИ Банка.

        """
        url = self._router.build_url("GET_CROSS_RATES")
        async for response in self._requests.fast().fetch(
                url=url,
                method='GET'
        ):
            return api_helper.simple_multiply_parse(
                lst_of_objects=response.response_data.get("result"),
                model=CrossRate
            )

    async def payment_by_payment_details(
            self,
            payment_sum: Sum,
            payment_method: PaymentMethod,
            fields: FreePaymentDetailsFields,
            payment_id: Optional[str] = None
    ) -> PaymentInfo:
        """
        Оплата услуг коммерческих организаций по их банковским реквизитам.

        :param payment_id: id платежа, если не передается, используется
         uuid4
        :param payment_sum: обьект Sum, в котором указывается сумма платежа
        :param payment_method: метод платежа
        :param fields: Набор реквизитов платежа
        """
        url = self._router.build_url("SPECIAL_PAYMENT")
        headers = deepcopy(self._router.config.DEFAULT_QIWI_HEADERS)
        payload = {
            "id": payment_id if isinstance(payment_id, str) else str(
                uuid.uuid4()
            ),
            "sum": payment_sum.dict(),
            "paymentMethod": payment_method.dict(),
            "fields": fields.dict()
        }
        async for response in self._requests.fast().fetch(
                url=url,
                json=payload,
                headers=headers
        ):
            return PaymentInfo.parse_obj(response.response_data)



import abc
import asyncio
import unittest
from typing import AsyncGenerator, Optional, Dict, Any, Union, List, Tuple

import aiohttp
from aiohttp import ClientSession, web
from aiohttp.typedefs import LooseCookies

from glQiwiApi import types
from glQiwiApi.core.web_hooks.dispatcher import Dispatcher
from glQiwiApi.types import Response


class SingletonABCMeta(abc.ABCMeta):
    """
    Abstract singleton metaclass, using for routers because in class methods
    it's not possible to get the router object,
    so we need singleton to get the same instances of routers

    """
    _instances: Dict[Any, Any] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonABCMeta, cls).__call__(
                *args, **kwargs
            )
        return cls._instances[cls]


class BaseStorage(abc.ABC):
    """
    Абстрактный класс контроллера кэша

    """

    @abc.abstractmethod
    def get_current(self, key: str) -> Any:
        raise NotImplementedError

    @abc.abstractmethod
    def clear(self, key: str, force: bool = False) -> Any:
        raise NotImplementedError

    def update_data(self, obj_to_cache: Any, key: Any) -> None:
        raise NotImplementedError

    def validate(self, kwargs: Dict[str, Any]) -> bool:
        """ Should validate some data from cache """


class AbstractRouter(metaclass=SingletonABCMeta):
    """Abstract class-router for building urls"""

    def __init__(self):
        self.config = self.setup_config()

    @abc.abstractmethod
    def setup_config(self) -> Any:
        raise NotImplementedError

    @staticmethod
    def _format_url_kwargs(url_: str, **kwargs: Any) -> Optional[str]:
        try:
            return url_.format(**kwargs)
        except KeyError:
            raise ValueError("Bad kwargs for url build")

    @abc.abstractmethod
    def build_url(self, api_method: str, **kwargs: Any) -> str:
        """Implementation of building url"""
        raise NotImplementedError


class AioTestCase(unittest.TestCase):
    """ Test case, but redone for asyncio """

    # noinspection PyPep8Naming
    def __init__(self, methodName='runTest', loop=None):
        self.loop = loop or asyncio.get_event_loop()
        self._function_cache = {}
        super(AioTestCase, self).__init__(methodName=methodName)

    def coroutine_function_decorator(self, func):
        def wrapper(*args, **kw):
            return self.loop.run_until_complete(func(*args, **kw))

        return wrapper

    def __getattribute__(self, item):
        attr = object.__getattribute__(self, item)
        if asyncio.iscoroutinefunction(attr):
            if item not in self._function_cache:
                self._function_cache[
                    item
                ] = self.coroutine_function_decorator(attr)
            return self._function_cache[item]
        return attr


class AbstractParser(abc.ABC):
    """ Abstract class of parser for send request to different API's"""
    session: Optional[aiohttp.ClientSession] = None

    @abc.abstractmethod
    async def _request(
            self,
            url: str,
            get_json: bool,
            method: str,
            set_timeout: bool,
            cookies: Optional[LooseCookies],
            json: Optional[dict],
            data: Optional[Dict[str, Union[
                str, int, List[
                    Union[str, int]
                ]]]
            ],
            headers: Optional[Dict[str, Union[str, int]]],
            params: Optional[
                Dict[str, Union[str, int, List[
                    Union[str, int]
                ]]]
            ]) -> Response:
        raise NotImplementedError

    @abc.abstractmethod
    async def fetch(
            self,
            *,
            times: int = 1,
            **kwargs
    ) -> AsyncGenerator[Response, None]:
        """ You can override it, but is must to return an AsyncGenerator """
        raise NotImplementedError

    def raise_exception(
            self,
            status_code: str,
            json_info: Optional[Dict[str, Any]] = None,
            message: Optional[str] = None
    ) -> None:
        """Метод для обработки исключений и лучшего логирования"""

    def create_session(self, **kwargs) -> None:
        """ Creating new session if old was close or it's None """
        if self.session is None:
            self.session = ClientSession(**kwargs)
        elif isinstance(self.session, ClientSession):
            if self.session.closed:
                self.session = ClientSession(**kwargs)


class BaseWebHookView(web.View):
    """
    Base webhook view for processing events
    You can make own view and than use it in code,
    just inheriting this base class

    """
    app_key_check_ip: Optional[str] = None
    """app_key_check_ip stores key to a storage"""

    app_key_handler_manager: Optional[str] = None
    """app_key_handler_manager"""

    @abc.abstractmethod
    def _check_ip(self, ip_address: str) -> bool:
        """_check_ip checks if given IP is in set of allowed ones"""
        raise NotImplementedError

    async def parse_update(self):
        """parse_update method that deals with marshaling json"""
        raise NotImplementedError

    def validate_ip(self):
        """ validating request ip address """
        if self.request.app.get(self.app_key_check_ip):
            ip_addr_data = self.check_ip()
            if not ip_addr_data[1]:
                raise web.HTTPUnauthorized()

    def check_ip(self) -> Union[Tuple[str, bool], Tuple[None, bool]]:
        """ checking ip, using headers or peer_name """
        forwarded_for = self.request.headers.get("X-Forwarded-For")
        if forwarded_for:
            return forwarded_for, self._check_ip(forwarded_for)
        peer_name = self.request.transport.get_extra_info("peername")

        if peer_name is not None:
            host, _ = peer_name
            return host, self._check_ip(host)

        return None, False

    @abc.abstractmethod
    async def post(self) -> Any:
        """
        Process POST request with basic IP validation.

        """
        self.validate_ip()

        update = await self.parse_update()

        self._hash_validator(update)

        await self.handler_manager.process_event(update)

    def _hash_validator(self, update: Union[
        types.Notification, types.WebHook
    ]) -> None:
        """ Validating by hash of update """

    @property
    def handler_manager(self) -> "Dispatcher":
        """ Return handler manager """
        return self.request.app.get(  # type: ignore
            self.app_key_handler_manager  # type: ignore
        )  # type: ignore



import uuid
from copy import deepcopy
from datetime import datetime
from typing import Union, Optional, Dict, List, Any, MutableMapping

from glQiwiApi.core import constants
from glQiwiApi.core.abstracts import AbstractRouter
from glQiwiApi.types import Bill, BillError, OptionalSum
from glQiwiApi.types.qiwi_types.bill import RefundBill, P2PKeys
from glQiwiApi.utils import basics as api_helper
from glQiwiApi.utils.exceptions import InvalidData


class QiwiKassaMixin:

    def __init__(
            self,
            p2p_router: AbstractRouter,
            default_router: AbstractRouter,
            requests_manager: Any,
            secret_p2p: str
    ):
        self._router = default_router
        self._p2p_router = p2p_router
        self._requests = requests_manager
        self.secret_p2p = secret_p2p

    def _auth_token(
            self,
            headers: MutableMapping,
            p2p: bool = False
    ) -> MutableMapping:
        ...

    async def reject_p2p_bill(self, bill_id: str) -> Bill:
        """
        Метод для отмены транзакции.

        :param bill_id: номер p2p транзакции
        :return: Bill obj
        """
        if not self.secret_p2p:
            raise InvalidData('Не задан p2p токен')
        data = deepcopy(self._p2p_router.config.P2P_DATA)
        headers = self._auth_token(data.headers, p2p=True)
        url = f'https://api.qiwi.com/partner/bill/v1/bills/{bill_id}/reject'
        async for response in self._requests.fast().fetch(
                url=url,
                method='POST',
                headers=headers
        ):
            return Bill.parse_obj(response.response_data)

    async def check_p2p_bill_status(self, bill_id: str) -> str:
        """
        Метод для проверки статуса p2p транзакции.\n
        Возможные типы транзакции: \n
        WAITING	Счет выставлен, ожидает оплаты	\n
        PAID	Счет оплачен	\n
        REJECTED	Счет отклонен\n
        EXPIRED	Время жизни счета истекло. Счет не оплачен\n
        Более подробная документация:
        https://developer.qiwi.com/ru/p2p-payments/?shell#invoice-status

        :param bill_id: номер p2p транзакции
        :return: статус транзакции строкой
        """
        if not self.secret_p2p:
            raise InvalidData('Не задан p2p токен')

        data = deepcopy(self._router.config.P2P_DATA)
        headers = self._auth_token(data.headers, p2p=True)
        url = self._p2p_router.build_url(
            "CHECK_P2P_BILL_STATUS",
            bill_id=bill_id
        )
        async for response in self._requests.fast().fetch(
                url=url,
                method='GET',
                headers=headers
        ):
            return Bill.parse_obj(response.response_data).status.value

    async def create_p2p_bill(
            self,
            amount: int,
            bill_id: Optional[str] = None,
            comment: Optional[str] = None,
            life_time: Optional[datetime] = None,
            theme_code: Optional[str] = None,
            pay_source_filter: Optional[List[str]] = None
    ) -> Union[Bill, BillError]:
        """
        Метод для выставление p2p счёта.
        Надежный способ для интеграции.
        Параметры передаются server2server с использованием авторизации.
        Возможные значения pay_source_filter:
          - 'qw'
          - 'card'
          - 'mobile'

        :param amount: сумма платежа
        :param bill_id: уникальный номер транзакции, если не передан,
         генерируется автоматически,
        :param life_time: дата, до которой счет будет доступен для оплаты.
        :param comment: комментарий к платежу
        :param theme_code: специальный код темы
        :param pay_source_filter: При открытии формы будут отображаться
         только указанные способы перевода
        """
        if not self.secret_p2p:
            raise InvalidData('Не задан p2p токен')

        if not isinstance(bill_id, (str, int)):
            bill_id = str(uuid.uuid4())

        _life_time = api_helper.datetime_to_str_in_iso(
            constants.DEFAULT_BILL_TIME if not life_time else life_time
        )

        data = deepcopy(self._p2p_router.config.P2P_DATA)

        headers = self._auth_token(data.headers, p2p=True)

        payload = api_helper.set_data_p2p_create(
            wrapped_data=data,
            amount=amount,
            comment=comment,
            theme_code=theme_code,
            pay_source_filter=pay_source_filter,
            life_time=str(_life_time)
        )
        url = self._p2p_router.build_url(
            "CREATE_P2P_BILL",
            bill_id=bill_id
        )
        async for response in self._requests.fast().fetch(
                url=url,
                json=payload,
                headers=headers,
                method='PUT'
        ):
            return Bill.parse_obj(response.response_data).initialize(self)

    async def get_bills(self, rows: int) -> List[Bill]:
        """
        Метод получения списка неоплаченных счетов вашего кошелька.
        Список строится в обратном хронологическом порядке.
        По умолчанию, список разбивается на страницы по 50 элементов в каждой,
        но вы можете задать другое количество элементов (не более 50).
        В запросе можно использовать фильтры по времени выставления счета,
        начальному идентификатору счета.
        """
        headers = self._auth_token(
            deepcopy(self._router.config.DEFAULT_QIWI_HEADERS)
        )
        if rows > 50:
            raise InvalidData('Можно получить не более 50 счетов')

        params = {
            'rows': rows,
            'statuses': 'READY_FOR_PAY'
        }
        async for response in self._requests.fast().fetch(
                url=self._router.build_url("GET_BILLS"),
                headers=headers,
                method='GET',
                params=params
        ):
            return api_helper.simple_multiply_parse(
                response.response_data.get("bills"), Bill
            )

    async def refund_bill(
            self,
            bill_id: Union[str, int],
            refund_id: Union[str, int],
            json_bill_data: Union[OptionalSum, Dict[str, Union[str, int]]]
    ) -> RefundBill:
        """
        Метод позволяет сделать возврат средств по оплаченному счету.
        в JSON-теле запроса параметра json_bill_data:\n
         amount.value - сумма возврата. \n
         amount.currency - валюта возврата.
        Может быть словарем или объектом OptionalSum\n
         Пример словаря: {
        "amount": {
            "currency": "RUB",
            "value": 1
            }
        }

        :param bill_id: уникальный идентификатор счета в системе мерчанта
        :param refund_id: уникальный идентификатор возврата в системе мерчанта.
        :param json_bill_data:
        :return: RefundBill object
        """
        url = self._router.build_url(
            "REFUND_BILL",
            refund_id=refund_id,
            bill_id=bill_id
        )
        headers = self._auth_token(
            deepcopy(self._router.config.DEFAULT_QIWI_HEADERS), p2p=True
        )
        async for response in self._requests.fast().fetch(
                url=url,
                headers=headers,
                method='PUT',
                json=json_bill_data if isinstance(
                    json_bill_data,
                    dict
                ) else json_bill_data.json()
        ):
            return RefundBill.parse_obj(response.response_data)

    async def create_p2p_keys(
            self,
            key_pair_name: str,
            server_notification_url: Optional[str] = None) -> P2PKeys:
        """
        Метод создает новые p2p ключи

        :param key_pair_name: Название пары токенов P2P (произвольная строка)
        :param server_notification_url: url для вебхуков, необязательный
         параметр
        """
        url = self._router.build_url("CREATE_P2P_KEYS")
        headers = self._auth_token(
            deepcopy(self._router.config.DEFAULT_QIWI_HEADERS), p2p=True
        )
        data = {
            'keysPairName': key_pair_name,
            'serverNotificationsUrl': server_notification_url
        }
        async for response in self._requests.fast().fetch(
                url=url,
                headers=headers,
                json=data
        ):
            return P2PKeys.parse_obj(response.response_data)



from datetime import datetime
from typing import Optional

from pydantic import BaseModel, Field

from glQiwiApi.types import Sum


class Payment(BaseModel):
    """ Scheme of webhook payment object """
    account: str = Field(..., alias="account")
    comment: str = Field(..., alias="comment")
    date: datetime = Field(..., alias="date")
    error_code: str = Field(..., alias="errorCode")
    person_id: int = Field(..., alias="personId")
    provider: int = Field(..., alias="provider")
    sign_fields: str = Field(..., alias="signFields")
    status: str = Field(..., alias="status")
    txn_id: str = Field(..., alias="txnId")
    type: str = Field(..., alias="type")
    commission: Optional[Sum] = Field(default=None, alias="commission")
    sum: Sum = Field(..., alias="sum")
    total: Sum = Field(..., alias="total")


class WebHook(BaseModel):
    """
    Хуки или уведомления с данными о событии (платеже/пополнении)

    """

    hash: Optional[str] = Field(default=None, alias="hash")
    hook_id: str = Field(..., alias="hookId")
    message_id: Optional[str] = Field(default=None, alias="messageId")
    test: bool = Field(..., alias="test")
    version: str = Field(..., alias="version")
    payment: Optional[Payment] = Field(default=None, alias="payment")


class HookParameters(BaseModel):
    """hookParameters object"""

    url: str = Field(..., alias="url")


class WebHookConfig(BaseModel):
    """WebHookConfig object"""

    hook_id: str = Field(..., alias="hookId")
    hook_type: str = Field(..., alias="hookType")
    txn_type: str = Field(..., alias="txnType")
    hook_parameters: HookParameters = Field(..., alias="hookParameters")


__all__ = ('WebHookConfig', 'WebHook')



from typing import Optional

from pydantic import BaseModel, Field, validator

from glQiwiApi.types.basics import Sum, Type
from glQiwiApi.types.qiwi_types.currency_parsed import CurrencyModel
from glQiwiApi.utils.currency_util import Currency


class Account(BaseModel):
    """ object: Account """
    alias: str
    title: str
    fs_alias: str = Field(alias="fsAlias")
    bank_alias: str = Field(alias="bankAlias")
    has_balance: bool = Field(alias="hasBalance")
    balance: Optional[Sum] = Field(const=None)
    currency: CurrencyModel
    account_type: Optional[Type] = Field(None, alias="type")
    is_default_account: bool = Field(alias="defaultAccount")

    @validator("currency", pre=True, check_fields=True)
    def humanize_pay_currency(cls, v):
        if not isinstance(v, int):
            return v
        return Currency.get(str(v))


__all__ = (
    'Account'
)



from .kassa_mixin import QiwiKassaMixin
from .master_mixin import QiwiMasterMixin
from .polling_mixin import HistoryPollingMixin
from .webhook_mixin import QiwiWebHookMixin
from .payment_mixin import QiwiPaymentsMixin

__all__ = (
    'QiwiKassaMixin',
    'QiwiMasterMixin',
    'HistoryPollingMixin',
    'QiwiWebHookMixin',
    'QiwiPaymentsMixin'
)



import pytest
from glQiwiApi.utils.currency_util import Currency, cur

pytestmark = pytest.mark.asyncio

_ = Currency()


async def test_currency_parser():
    from glQiwiApi.types.qiwi_types.currency_parsed import CurrencyModel
    condition = all(
        isinstance(_.get(key), CurrencyModel) for key in cur.described.keys()
    )
    assert condition



import pytest
from glQiwiApi import types, QiwiMaps

pytestmark = pytest.mark.asyncio


@pytest.fixture(name="data")
def maps_data():
    polygon = types.Polygon((55.690881, 37.386282), (55.580184, 37.826078))
    yield {
        "polygon": polygon,
        "zoom": 12,
        "cache_terminals": True
    }


@pytest.fixture(name="maps")
async def maps_fixture():
    _maps = QiwiMaps()
    yield _maps
    await _maps.close()


async def test_terminals(maps: QiwiMaps, data: dict):
    async with maps:
        result = await maps.terminals(**data)
    assert all(isinstance(t, types.Terminal) for t in result)


async def test_partners(maps: QiwiMaps):
    async with maps:
        result = await maps.partners()
    assert all(isinstance(p, types.Partner) for p in result)



from dataclasses import dataclass
from http.cookies import SimpleCookie
from typing import Optional, Union, Dict, Any

from aiohttp.typedefs import RawHeaders

from glQiwiApi.utils.exceptions import ProxyError


@dataclass
class Response:
    """ object: Response """
    status_code: int
    response_data: Any = None
    url: Optional[str] = None
    raw_headers: Optional[RawHeaders] = None
    cookies: Optional[SimpleCookie] = None
    ok: bool = False
    content_type: Optional[str] = None
    host: Optional[str] = None

    @classmethod
    def bad_response(cls) -> 'Response':
        return cls(
            status_code=500,
            response_data=ProxyError
        )

    def __str__(self) -> str:
        return f"{self.status_code} from {self.url}. OK: {self.ok}"


@dataclass
class WrapperData:
    """ object: WrapperData """
    headers: Optional[Dict[str, Union[str, int]]]
    data: Optional[Dict[str, Union[str, Dict[Any, Any]]]] = None
    json: Optional[Dict[str, Union[str, Dict[Any, Any]]]] = None
    cookies: Optional[Dict[str, Union[str, int]]] = None


__all__ = ('Response', 'WrapperData')



import asyncio
import copy
import logging
from typing import Optional, Any, Awaitable, Dict, Tuple, Callable, \
    MutableMapping

from aiohttp import web

from glQiwiApi.core.abstracts import AbstractRouter
from glQiwiApi.core.web_hooks import dispatcher, server
from glQiwiApi.core.web_hooks.config import Path
from glQiwiApi.types import WebHookConfig
from glQiwiApi.utils import basics as api_helper
from glQiwiApi.utils.exceptions import RequestError


class QiwiWebHookMixin:
    """Mixin, which implements webhook logic"""

    def __init__(
            self,
            router: AbstractRouter,
            requests_manager: Any,
            secret_p2p: str
    ):
        self.dispatcher = dispatcher.Dispatcher(
            loop=asyncio.get_event_loop(),
            wallet=self
        )
        self._router = router
        self._requests = requests_manager
        self.secret_p2p = secret_p2p

    def _auth_token(
            self,
            headers: MutableMapping,
            p2p: bool = False
    ) -> MutableMapping:
        ...

    async def _register_webhook(
            self,
            web_url: Optional[str],
            txn_type: int = 2
    ) -> Optional[WebHookConfig]:
        """
        This method register a new webhook

        :param web_url: service url
        :param txn_type:  0 => incoming, 1 => outgoing, 2 => all
        :return: Active Hooks
        """
        url = self._router.build_url("REG_WEBHOOK")
        async for response in self._requests.fast().fetch(
                url=url,
                method='PUT',
                headers=self._auth_token(copy.deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                )),
                params={
                    'hookType': 1,
                    'param': web_url,
                    'txnType': txn_type
                }
        ):
            return WebHookConfig.parse_obj(response.response_data)

    async def get_current_webhook(self) -> Optional[WebHookConfig]:
        """
        Список действующих (активных) обработчиков уведомлений,
         связанных с вашим кошельком, можно получить данным запросом.
        Так как сейчас используется только один тип хука - webhook,
         то в ответе содержится только один объект данных

        """
        url = self._router.build_url("GET_CURRENT_WEBHOOK")
        async for response in self._requests.fast().fetch(
                url=url,
                method='GET',
                headers=self._auth_token(copy.deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                ))
        ):
            try:
                return WebHookConfig.parse_obj(response.response_data)
            except RequestError:
                return None

    async def _send_test_notification(self) -> Dict[str, str]:
        """
        Для проверки вашего обработчика webhooks используйте данный запрос.
        Тестовое уведомление отправляется на адрес, указанный при вызове
        register_webhook

        """
        url = self._router.build_url("SEND_TEST_NOTIFICATION")
        async for response in self._requests.fast().fetch(
                url=url,
                method='GET',
                headers=self._auth_token(copy.deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                ))
        ):
            return response.response_data

    async def get_webhook_secret_key(self, hook_id: str) -> str:
        """
        Каждое уведомление содержит цифровую подпись сообщения,
         зашифрованную ключом.
        Для получения ключа проверки подписи используйте данный запрос.

        :param hook_id: UUID of webhook
        :return: Base64-закодированный ключ
        """
        url = self._router.build_url(
            "GET_WEBHOOK_SECRET",
            hook_id=hook_id
        )
        async for response in self._requests.fast().fetch(
                url=url,
                method='GET',
                headers=self._auth_token(copy.deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                ))
        ):
            return response.response_data.get('key')

    async def delete_current_webhook(self) -> Optional[Dict[str, str]]:
        """
        Method to delete webhook

        :return: Описание результата операции
        """
        try:
            hook = await self.get_current_webhook()
        except RequestError as ex:
            raise RequestError(
                message=" You didn't register any webhook to delete ",
                status_code='422',
                json_info=ex.json()
            ) from None

        url = self._router.build_url(
            "DELETE_CURRENT_WEBHOOK",
            hook_id=hook.hook_id
        )
        async for response in self._requests.fast().fetch(
                url=url,
                headers=self._auth_token(copy.deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                )),
                method='DELETE'
        ):
            return response.response_data

    async def change_webhook_secret(self, hook_id: str) -> str:
        """
        Для смены ключа шифрования уведомлений используйте данный запрос.

        :param hook_id: UUID of webhook
        :return: Base64-закодированный ключ
        """
        url = self._router.build_url(
            "CHANGE_WEBHOOK_SECRET",
            hook_id=hook_id
        )
        async for response in self._requests.fast().fetch(
                url=url,
                headers=self._auth_token(copy.deepcopy(
                    self._router.config.DEFAULT_QIWI_HEADERS
                )),
                method='POST'
        ):
            return response.response_data.get('key')

    async def bind_webhook(
            self,
            url: Optional[str] = None,
            transactions_type: int = 2,
            *,
            send_test_notification: bool = False,
            delete_old: bool = False
    ) -> Tuple[Optional[WebHookConfig], str]:
        """
        [NON-API] EXCLUSIVE method to register new webhook or get old

        :param url: service url
        :param transactions_type: 0 => incoming, 1 => outgoing, 2 => all
        :param send_test_notification:  test_qiwi will send you test webhook update
        :param delete_old: boolean, if True - delete old webhook

        :return: Tuple of Hook and Base64-encoded key
        """
        key: Optional[str] = None

        if delete_old:
            await self.delete_current_webhook()

        try:
            # Try to register new webhook
            webhook = await self._register_webhook(
                web_url=url,
                txn_type=transactions_type
            )
        except (RequestError, TypeError):
            # Catching exception, if webhook already was registered
            try:
                webhook = await self.get_current_webhook()
            except RequestError as ex:
                raise RequestError(
                    message="You didn't pass on url to register new hook "
                            "and you didn't have registered webhooks",
                    status_code="422",
                    json_info=ex.json()
                )
            key = await self.get_webhook_secret_key(webhook.hook_id)
            return webhook, key

        if send_test_notification:
            await self._send_test_notification()

        if not isinstance(key, str):
            key = await self.get_webhook_secret_key(webhook.hook_id)

        return webhook, key

    def start_webhook(
            self,
            host: str = "localhost",
            port: int = 8080,
            path: Optional[Path] = None,
            app: Optional["web.Application"] = None,
            on_startup: Optional[
                Callable[
                    [web.Application], Awaitable[None]
                ]] = None,
            on_shutdown: Optional[
                Callable[
                    [web.Application], Awaitable[None]
                ]] = None,
            **logger_config: Any
    ):
        """
        Blocking function, which listening webhooks

        :param host: server host
        :param port: server port that open for tcp/ip trans.
        :param path: path for test_qiwi that will send requests
        :param app: pass web.Application
        :param on_startup: coroutine,which will be executed on startup
        :param on_shutdown: coroutine, which will be executed on shutdown
        """
        self._requests.without_context = True

        app = app if app is not None else web.Application()

        hook_config, key = api_helper.sync(self.bind_webhook)

        server.setup(
            dispatcher=self.dispatcher,
            app=app,
            path=Path() if not path else path,
            secret_key=self.secret_p2p,
            base64_key=key,
            on_startup=on_startup,
            on_shutdown=on_shutdown,
            instance=self
        )

        logging.basicConfig(**logger_config)

        web.run_app(app, host=host, port=port)

    @property
    def transaction_handler(self):
        """
        Handler manager for default test_qiwi transactions,
        you can pass on lambda filter, if you want,
        but it must to return a boolean

        """
        return self.dispatcher.transaction_handler_wrapper

    @property
    def bill_handler(self):
        """
        Handler manager for p2p bills,
        you can pass on lambda filter, if you want
        But it must to return a boolean

        """
        return self.dispatcher.bill_handler_wrapper



import datetime
import pathlib
import uuid
from typing import Dict, Union

import pytest
from glQiwiApi import types, InvalidData
from glQiwiApi import QiwiWrapper

pytestmark = pytest.mark.asyncio


@pytest.fixture(name='api')
async def api_fixture(credentials: Dict[str, str]):
    """ Api fixture """
    _wrapper = QiwiWrapper(**credentials)
    yield _wrapper
    await _wrapper.close()


async def test_get_balance(api: QiwiWrapper):
    from glQiwiApi.types.qiwi_types.currency_parsed import CurrencyModel
    async with api:
        result = await api.get_balance()
    assert isinstance(result, types.Sum)
    assert isinstance(result.currency, CurrencyModel)


@pytest.mark.parametrize(
    "payload",
    [
        {"rows_num": 50},
        {"rows_num": 50,
         "operation": "IN"},
        {
            "rows_num": 50,
            "operation": "IN",
            "start_date": datetime.datetime.now() - datetime.timedelta(
                days=50),
            "end_date": datetime.datetime.now()
        }
    ]
)
async def test_transactions(api: QiwiWrapper, payload: dict):
    async with api:
        result = await api.transactions(**payload)

    assert all(isinstance(t, types.Transaction) for t in result)


@pytest.mark.parametrize("payload", [
    {
        "transaction_id": 21601937643,
        "transaction_type": "OUT"
    }
])
async def test_transaction_info(api: QiwiWrapper, payload: dict):
    async with api:
        result = await api.transaction_info(**payload)

    assert isinstance(result, types.Transaction)


async def test_identification(api: QiwiWrapper):
    async with api:
        result = await api.identification
    assert isinstance(result, types.Identification)


@pytest.mark.parametrize("payload", [
    {
        "transaction_type": "OUT",
        "amount": 1
    },
    {
        "transaction_type": "OUT",
        "amount": 1,
        "sender": "+380985272064"
    },
    {
        "transaction_type": "OUT",
        "amount": 1,
        "sender": "+380985272064",
        "comment": "+comment+"
    }
])
async def test_check_transaction(api: QiwiWrapper, payload: dict):
    async with api:
        r = await api.check_transaction(**payload)
    assert r is True


async def test_get_limits(api: QiwiWrapper):
    async with api:
        result = await api.get_limits()

    assert isinstance(result, dict)
    assert all(isinstance(t, types.Limit) for t in result.values())


async def test_get_list_of_cards(api: QiwiWrapper):
    async with api:
        result = await api.get_list_of_cards()
    assert all(isinstance(c, types.Card) for c in result)


async def test_get_receipt_and_save(
        api: QiwiWrapper, path_to_dir: pathlib.Path
):
    from ..types.dataset import RECEIPT_FILE_NAME
    file_name = RECEIPT_FILE_NAME
    payload = {
        "transaction_id": 21601937643,
        "transaction_type": "OUT",
        "dir_path": path_to_dir,
    }
    async with api:
        await api.get_receipt(**payload, file_name=file_name)
    assert (path_to_dir / (file_name + ".pdf")).is_file()


async def test_account_info(api: QiwiWrapper):
    async with api:
        result = await api.account_info

    assert isinstance(result, types.QiwiAccountInfo)


@pytest.mark.parametrize("payload", [
    {
        "start_date": datetime.datetime.now() - datetime.timedelta(days=50),
        "end_date": datetime.datetime.now()
    },
    {
        "start_date": datetime.datetime.now() - datetime.timedelta(days=50),
        "end_date": datetime.datetime.now(),
        "operation": "IN"
    },
    {
        "start_date": datetime.datetime.now() - datetime.timedelta(days=50),
        "end_date": datetime.datetime.now(),
        "operation": "OUT"
    },
    {
        "start_date": datetime.datetime.now() - datetime.timedelta(days=50),
        "end_date": datetime.datetime.now(),
        "operation": "ALL"
    },
    {
        "start_date": datetime.datetime.now() - datetime.timedelta(days=50),
        "end_date": datetime.datetime.now(),
        "operation": "ALL",
        "sources": ["QW_RUB"]
    },
    {
        "start_date": datetime.datetime.now() - datetime.timedelta(days=50),
        "end_date": datetime.datetime.now(),
        "operation": "ALL",
        "sources": ["QW_RUB", "QW_EUR", "QW_USD"]
    }
])
async def test_fetch_statistic(api: QiwiWrapper, payload: dict):
    async with api:
        result = await api.fetch_statistics(**payload)

    assert isinstance(result, types.Statistic)


async def test_list_of_balances(api: QiwiWrapper):
    async with api:
        balances = await api.list_of_balances()

    assert all(isinstance(b, types.Account) for b in balances)


async def test_fail_fetch_statistic(api: QiwiWrapper):
    payload = {
        # Wrong start date
        "start_date": datetime.datetime.now() - datetime.timedelta(days=100),
        "end_date": datetime.datetime.now(),
        "operation": "OUT"
    }
    async with api:
        with pytest.raises(ValueError):
            await api.fetch_statistics(**payload)


@pytest.mark.parametrize(
    "payload",
    [
        {"amount": 1},
        {"amount": 1, "comment": "test_comment"},
        {
            "amount": 1,
            "comment": "test_comment",
            "life_time": datetime.datetime.now() + datetime.timedelta(hours=5)
        },
        {
            "amount": 1,
            "comment": "test_comment",
            "life_time": datetime.datetime.now() + datetime.timedelta(hours=5),
            "bill_id": str(uuid.uuid4())
        }
    ])
async def test_create_p2p_bill(api: QiwiWrapper, payload: dict):
    async with api:
        result = await api.create_p2p_bill(**payload)

    assert isinstance(result, types.Bill)
    assert payload["amount"] == result.amount.value


async def test_check_p2p_bill_status(api: QiwiWrapper):
    async with api:
        result = await api.check_p2p_bill_status(
            bill_id="8d517426-920a-4711-86c2-4267784bf901"
        )
    assert isinstance(result, str)


async def test_check_p2p_on_object(api: QiwiWrapper):
    async with api:
        bill = await api.create_p2p_bill(amount=1)
        result = await bill.check()

    assert isinstance(result, bool)


@pytest.mark.parametrize("rows", [5, 10, 50])
async def test_get_bills(api: QiwiWrapper, rows: int):
    async with api:
        result = await api.get_bills(rows=rows)

    assert isinstance(result, list)
    assert all(isinstance(b, types.Bill) for b in result)


async def test_check_restriction(api: QiwiWrapper):
    async with api:
        result = await api.check_restriction()

    assert isinstance(result, list)
    assert all(isinstance(r, types.Restriction) for r in result)


async def test_commission(api: QiwiWrapper):
    async with api:
        result = await api.commission(
            to_account="+380985272064",
            pay_sum=999
        )
    assert isinstance(result, types.Commission)


async def test_get_cross_rates(api: QiwiWrapper):
    async with api:
        result = await api.get_cross_rates()

    assert all(isinstance(r, types.CrossRate) for r in result)


class TestFail:

    @pytest.mark.parametrize("rows_num", [-5, 51, 0])
    async def test_transactions_fail(self, api: QiwiWrapper, rows_num: int):
        async with api:
            with pytest.raises(InvalidData):
                await api.transactions(rows_num=rows_num)

    @pytest.mark.parametrize("start_date,end_date", [
        {
            "start_date": datetime.datetime.now() - datetime.timedelta(
                days=100),
            "end_date": datetime.datetime.now()
        },
        {
            "start_date": datetime.datetime.now() - datetime.timedelta(
                days=80),
            "end_date": "Wrong Parameter"
        }
    ])
    async def test_fetch_statistic_fail(
            self,
            api: QiwiWrapper,
            start_date: Union[datetime.datetime, datetime.timedelta],
            end_date: Union[datetime.datetime, datetime.timedelta]
    ):
        async with api:
            with pytest.raises(ValueError):
                await api.fetch_statistics(
                    start_date=start_date,
                    end_date=end_date
                )



import asyncio
import datetime
from typing import Dict

import pytest

from glQiwiApi import YooMoneyAPI, types
from glQiwiApi.types import OperationType

pytestmark = pytest.mark.asyncio


@pytest.fixture(name='api')
async def api_fixture(yoo_credentials: Dict[str, str]):
    """ Api fixture """
    _wrapper = YooMoneyAPI(**yoo_credentials)
    yield _wrapper
    await _wrapper.close()


async def test_get_balance(api: YooMoneyAPI):
    async with api:
        balance = await api.balance

    assert isinstance(balance, float)


@pytest.mark.parametrize("payload", [
    {},
    {
        "records": 10,
    },
    {
        "records": 50,
        "operation_types": [OperationType.PAYMENT, OperationType.DEPOSITION]
    },
    {
        "records": 40,
        "operation_types": [OperationType.PAYMENT]
    },
    {
        "records": 80,
        "operation_types": [OperationType.PAYMENT, OperationType.DEPOSITION],
        "start_date": datetime.datetime.now() - datetime.timedelta(days=80),
        "end_date": datetime.datetime.now()
    }
])
async def test_get_transactions(api: YooMoneyAPI, payload: dict):
    async with api:
        transactions = await api.transactions(**payload)

    assert all(isinstance(txn, types.Operation) for txn in transactions)


async def test_account_info(api: YooMoneyAPI):
    async with api:
        info = await api.account_info

    assert isinstance(info, types.AccountInfo)


@pytest.mark.parametrize("operation_id", [
    "672180330623679897", "671568515431002412"
])
async def test_get_transaction_info(api: YooMoneyAPI, operation_id: str):
    async with api:
        transaction = await api.transaction_info(operation_id)
    assert isinstance(transaction, types.OperationDetails)


@pytest.mark.parametrize("payload", [
    {"amount": 2},
    {
        "amount": 2,
        "transaction_type": "out"
    },
    {
        "amount": 2,
        "transaction_type": "out",
        "rows_num": 20
    },
    {
        "amount": 2,
        "transaction_type": "out",
        "rows_num": 20,
        "recipient": "4100116633099701"
    }
])
async def test_check_transaction(api: YooMoneyAPI, payload: dict):
    async with api:
        result = await api.check_transaction(**payload)

    assert isinstance(result, bool)


async def test_send_and_check_txn(api: YooMoneyAPI):
    async with api:
        payload = {
            "amount": 2,
            "comment": "unit_test"
        }
        await api.send(
            to_account="4100116633099701",
            **payload
        )
        answer = await api.check_transaction(
            **payload,
            transaction_type="out"
        )
    assert answer is True



import asyncio
import base64
import binascii
import concurrent.futures as futures
import datetime
import functools as ft
import hashlib
import hmac
import pathlib
import re
import time
from contextvars import ContextVar
from copy import deepcopy

import aiofiles
import pytz
from pydantic import ValidationError, BaseModel
from pytz.reference import LocalTimezone

try:
    import orjson
except (ModuleNotFoundError, ImportError):
    import json as orjson

# Gets your local time zone
Local = LocalTimezone()


def measure_time(func):
    """
    Декоратор для замера времени выполнения функции

    :param func:
    """

    @ft.wraps(func)
    async def wrapper(*args, **kwargs):
        start_time = time.monotonic()
        result = await func(*args, **kwargs)
        execute_time = time.monotonic() - start_time
        print(
            f'Function `{func.__name__}` executed for {execute_time} secs')
        return result

    return wrapper


def datetime_to_str_in_iso(obj, yoo_money_format=False):
    """
    Converts a date to a standard format for API's

    :param obj: datetime object to parse to string
    :param yoo_money_format: boolean
    :return: string - parsed date
    """
    if not isinstance(obj, datetime.datetime):
        return ''
    if yoo_money_format:
        # Приводим время к UTC,
        # так как yoomoney апи принимает именно в таком формате
        return pytz.utc.localize(obj).replace(tzinfo=None).isoformat(
            ' ').replace(" ", "T") + "Z"
    local_date = str(datetime.datetime.now(tz=Local))
    pattern = re.compile(r'[+]\d{2}[:]\d{2}')
    from_pattern = re.findall(pattern, local_date)[0]
    return obj.isoformat(' ').replace(" ", "T") + from_pattern


def parse_auth_link(response_data):
    """
    Parse link for getting code, which needs to be entered in the method
    get_access_token

    :param response_data:
    """
    regexp = re.compile(
        r'https://yoomoney.ru/oauth2/authorize[?]requestid[=]\w+'
    )
    return re.findall(regexp, str(response_data))[0]


def check_dates(start_date, end_date, payload_data):
    """ Check correctness of transferred dates and add it to request """
    if isinstance(
            start_date, (datetime.datetime, datetime.timedelta)
    ) and isinstance(
        end_date, (datetime.datetime, datetime.timedelta)
    ):
        if (end_date - start_date).total_seconds() > 0:
            payload_data.update(
                {
                    'startDate': datetime_to_str_in_iso(
                        start_date
                    ),
                    'endDate': datetime_to_str_in_iso(
                        end_date
                    )
                }
            )
        else:
            raise Exception(
                'end_date не может быть больше чем start_date'
            )
    return payload_data


def parse_commission_request_payload(
        default_data,
        auth_maker,
        pay_sum,
        to_account
):
    """Set commission payload"""
    payload = deepcopy(default_data)
    payload.headers = auth_maker(payload.headers)
    payload.json['purchaseTotals']['total']['amount'] = pay_sum
    payload.json['account'] = to_account
    return payload, "99" if len(to_account) <= 15 else None


def parse_card_data(
        default_data,
        trans_sum,
        to_card,
        auth_maker,
):
    """Set card data payload"""
    data = deepcopy(default_data)
    data.json['sum']['amount'] = trans_sum
    data.json['fields']['account'] = to_card
    data.headers = auth_maker(headers=data.headers)
    return data


def parse_headers(content_json=False, auth=False):
    """
    Функция для добавления некоторых заголовков в запрос
    """
    headers = {
        'Host': 'yoomoney.ru',
        'Content-Type': 'application/x-www-form-urlencoded'
    }
    if content_json:
        headers.update(
            {'Accept': 'application/json'}
        )
    if auth:
        headers.update(
            {'Authorization': 'Bearer {token}'}
        )
    return headers


def set_data_to_wallet(data, to_number, trans_sum, comment, currency):
    """
    Setting data for "wallet to wallet" transfer

    :param data:
    :param trans_sum:
    :param to_number:
    :param comment:
    :param currency:
    """
    data.json['sum']['amount'] = str(trans_sum)
    data.json['sum']['currency'] = currency
    data.json['fields']['account'] = to_number
    data.json['comment'] = comment
    data.headers.update({'User-Agent': 'Android v3.2.0 MKT'})
    return data


def set_data_p2p_create(wrapped_data, amount, life_time, comment, theme_code,
                        pay_source_filter):
    """
    Setting data for p2p form creation transfer

    :param wrapped_data:
    :param amount:
    :param life_time:
    :param comment:
    :param theme_code:
    :param pay_source_filter:
    """
    wrapped_data.json['amount']['value'] = str(amount)
    wrapped_data.json['comment'] = comment
    wrapped_data.json['expirationDateTime'] = life_time
    if isinstance(pay_source_filter, list):
        source_filters = " ".join(pay_source_filter).replace(" ", ",")
        wrapped_data.json['customFields'][
            'paySourcesFilter'
        ] = source_filters
    if isinstance(theme_code, str):
        wrapped_data.json['customFields']['themeCode'] = theme_code
    if not isinstance(theme_code, str) and pay_source_filter not in [
        'qw',
        'card',
        'mobile'
    ]:
        wrapped_data.json.pop('customFields')
    return wrapped_data.json


#
def multiply_objects_parse(lst_of_objects, model):
    """
    Function to handle list of objects

    :param lst_of_objects: usually its response.response_data
    :param model: pydantic model, which will parse data
    """
    objects = []
    for obj in lst_of_objects:
        try:
            if isinstance(obj, dict):
                obj: str
                objects.append(model.parse_obj(obj))
                continue
            for detached_obj in obj:
                objects.append(model.parse_obj(detached_obj))
        except ValidationError as ex:
            print(ex.json(indent=4))
    return objects


def simple_multiply_parse(lst_of_objects, model):
    """
    Parse simple objects, which cant raise ValidationError

    :param lst_of_objects: usually its response.response_data
    :param model: pydantic model, which will parse data
    """
    objects = []
    for obj in lst_of_objects:
        objects.append(model.parse_obj(obj))
    return objects


def take_event_loop(set_debug: bool = False):
    try:
        loop = asyncio.get_running_loop()
        if loop.is_closed():
            raise RuntimeError()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    loop.set_debug(set_debug)
    return loop


def hmac_key(key, amount, status, bill_id, site_id):
    """
    Функция расшифровки подписи webhook

    :param key: ключ webhook, закодированный в Base64
    :param amount: сумма p2p платежа
    :param status: статус платежа
    :param bill_id: unique p2p id
    :param site_id:
    """
    byte_key = binascii.unhexlify(key)
    invoice_params = (
        f"{amount.currency}|{amount.value}|{bill_id}|{site_id}|{status.value}"
    ).encode("utf-8")

    return hmac.new(
        byte_key, invoice_params, hashlib.sha256
    ).hexdigest().upper()


def hmac_for_transaction(
        webhook_key_base64,
        amount,
        txn_type,
        account,
        txn_id,
        txn_hash
):
    invoice_params = (
        f"{amount.currency}|{amount.amount}|{txn_type}|{account}|{txn_id}"
    )
    webhook_key = base64.b64decode(bytes(webhook_key_base64, 'utf-8'))
    return hmac.new(webhook_key, invoice_params.encode('utf-8'),
                    hashlib.sha256).hexdigest() == txn_hash


def custom_load(data):
    """
    Custom loads for each pydantic model, because
    it guard API from different errors

    :param data: class data
    """
    return orjson.loads(orjson.dumps(data))


class Parser(BaseModel):
    """ Модель pydantic для перевода строки в datetime """
    dt: datetime.datetime


def allow_response_code(status_code):
    """
    Декоратор, который позволяет разрешить определенный код ответа от апи

    :param status_code: статус код, который будет рассмотрен как правильный,
     кроме 200
    """

    def wrap_func(func):
        async def wrapper(*args, **kwargs):
            from glQiwiApi import RequestError
            try:
                await func(*args, **kwargs)
            except RequestError as error:
                if error.status_code == str(status_code):
                    info = error.json_info
                    return {"success": True} if not info else info
                return {"success": False}

        return wrapper

    return wrap_func


def qiwi_master_data(ph_number, data):
    payload = deepcopy(data)
    payload["fields"]["account"] = ph_number
    return payload


def to_datetime(string_representation):
    """
    Вспомогательная функция для перевода строки во время

    :param string_representation: дата в виде строки
    :return: datetime representation
    """
    try:
        parsed = orjson.dumps(
            {'dt': string_representation}
        )
        return Parser.parse_obj(parsed).dt
    except (ValidationError, orjson.JSONDecodeError) as ex:
        return ex.json(indent=4)


def new_card_data(ph_number, order_id, data):
    payload = deepcopy(data)
    url = "https://edge.qiwi.com" + "/sinap/test_api/v2/terms/32064/payments"
    payload["fields"].pop("vas_alias")
    payload["fields"].update(order_id=order_id)
    payload["fields"]["account"] = ph_number
    return url, payload


def sync_measure_time(func):
    """
    Decorator, which measures time your synchronous functions

    :param func:
    """

    @ft.wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.monotonic()
        result = func(*args, **kwargs)
        execute_time = time.monotonic() - start_time
        print(
            f'Function `{func.__name__}` executed for {execute_time} secs')
        return result

    return wrapper


def parse_amount(txn_type, txn):
    amount = txn.amount if txn_type == 'in' else txn.amount_due
    comment = txn.comment if txn_type == 'in' else txn.message
    return amount, comment


def check_params(amount_, amount, txn, transaction_type):
    if amount is not None:
        if amount_ >= amount:
            if txn.direction == transaction_type:
                return True
    return False


def run_forever_safe(loop) -> None:
    """ run a loop for ever and clean up after being stopped """

    loop.run_forever()
    # NOTE: loop.run_forever returns after calling loop.stop

    safe_cancel(loop=loop)


def safe_cancel(loop) -> None:
    # -- cancel all tasks and close the loop gracefully

    loop_tasks_all = asyncio.all_tasks(loop=loop)

    for task in loop_tasks_all:
        task.cancel()
    # NOTE: `cancel` does not guarantee that the Task will be cancelled

    for task in loop_tasks_all:
        if not (task.done() or task.cancelled()):
            try:
                # wait for task cancellations
                loop.run_until_complete(task)
            except asyncio.CancelledError:
                pass
    # Finally, close event loop
    loop.close()


def _await_sync(future):
    """ synchronously waits for a task """
    return future.result()


def _cancel_future(loop, future, executor) -> None:
    """ cancels future if any exception occurred """
    executor.submit(loop.call_soon_threadsafe, future.cancel)


def _stop_loop(loop) -> None:
    """ stops an event loop """
    loop.stop()


class AdaptiveExecutor(futures.ThreadPoolExecutor):
    """ object: AdaptiveExecutor """

    def __init__(self, max_workers=None, **kwargs):
        super().__init__(max_workers, 'sync_adapter_', **kwargs)
        self.max_workers = max_workers
        self.is_from_running_loop = ContextVar('Adapter_', default=False)


def _construct_all():
    """ Get or create new event loop """
    loop = take_event_loop()
    executor = AdaptiveExecutor()
    loop.set_default_executor(executor)
    return loop, executor


def _on_shutdown(executor, loop):
    """ Do some cleanup """
    if not executor.is_from_running_loop.get():
        loop.call_soon_threadsafe(_stop_loop, loop)
    executor.shutdown(wait=True)


def sync(func, *args, **kwargs):
    """
    Function to use async functions(libraries) synchronously

    :param func: Async function, which you want to execute in synchronous code
    :param args: args, which need your async func
    :param kwargs: kwargs, which need your async func
    """

    # construct an event loop and executor
    loop, executor = _construct_all()
    # Run our coroutine thread safe
    wrapped_future = asyncio.run_coroutine_threadsafe(
        func(*args, **kwargs),
        loop=loop
    )
    # Run loop.run_forever(), but do it safely
    executor.submit(run_forever_safe, loop)
    try:
        # Get result
        return _await_sync(wrapped_future)
    finally:
        # Cleanup
        _on_shutdown(executor, loop)


def check_transaction(
        transactions,
        amount,
        transaction_type,
        sender,
        comment
):
    for txn in transactions:
        if float(txn.sum.amount) >= amount:
            if txn.type == transaction_type:
                if txn.comment == comment and txn.to_account == sender:
                    return True
                elif comment and sender:
                    continue
                elif txn.to_account == sender:
                    return True
                elif sender:
                    continue
                elif txn.comment == comment:
                    return True
    return False


def parse_limits(response, model):
    limits = {}
    for code, limit in response.response_data.get("limits").items():
        if not limit:
            continue
        limits.update({code: model.parse_obj(limit[0])})
    return limits


def override_error_messages(status_codes):
    """
    Function, which helps to override error message

    :param status_codes: special dictionary
    """

    def functor(func):
        @ft.wraps(func)
        async def wrapper(*args, **kwargs):
            from glQiwiApi import RequestError
            try:
                return await func(*args, **kwargs)
            except RequestError as ex:
                if int(ex.status_code) in status_codes.keys():
                    error = status_codes.get(int(ex.status_code))
                    ex = RequestError(
                        message=error.get("message"),
                        status_code=ex.status_code,
                        json_info=error.get("json_info"),
                        additional_info=ex.additional_info
                    )
                raise ex

        return wrapper

    return functor


def check_api_method(api_method):
    if not isinstance(api_method, str):
        raise ValueError("Invalid type of api_method(must  be string)")


def check_dates_for_statistic_request(start_date, end_date):
    first_expression = isinstance(
        start_date, datetime.datetime
    ) and isinstance(end_date, datetime.datetime)
    second_expression = isinstance(
        start_date, datetime.timedelta
    ) and isinstance(end_date, datetime.timedelta)
    if first_expression or second_expression:
        delta: datetime.timedelta = end_date - start_date
        if delta.days > 90:
            raise ValueError(
                'Максимальный период для выгрузки'
                ' статистики - 90 календарных дней.'
            )
    else:
        raise ValueError(
            'Вы передали значения начальной '
            'и конечной даты в неправильном формате.'
        )


async def save_file(dir_path, file_name, data):
    if isinstance(dir_path, str):
        dir_path = pathlib.Path(dir_path)
    if not dir_path.is_dir():
        raise RuntimeError("Invalid path to save, its not directory!")
    dir_path: pathlib.Path
    path_to_file = dir_path / (file_name + '.pdf')
    async with aiofiles.open(path_to_file, 'wb') as file:
        return await file.write(data)



import asyncio
import datetime
from typing import Any, Union, Optional, List, \
    Tuple, Callable

from aiohttp import ClientTimeout

from glQiwiApi.types.qiwi_types.transaction import Transaction
from glQiwiApi.utils import basics as api_helper
from glQiwiApi.utils.exceptions import NoUpdatesToExecute

DEFAULT_TIMEOUT = 20.0  # timeout in seconds


def _get_last_payment(
        history: List[Transaction]
) -> Tuple[Optional[Transaction], Optional[int]]:
    """
    Function, which gets last payment and his id of history

    :param history:
    """
    last_payment = history[0]
    last_payment_id = last_payment.transaction_id
    return last_payment, last_payment_id


def _setup_callbacks(
        dispatcher,
        on_startup: Optional[Callable] = None,
        on_shutdown: Optional[Callable] = None
):
    """
    Function, which setup callbacks and set it to dispatcher object

    :param on_startup:
    :param on_shutdown:
    """
    if on_startup is not None:
        dispatcher["on_startup"] = on_startup
    if on_shutdown is not None:
        dispatcher["on_shutdown"] = on_shutdown


def parse_timeout(
        timeout: Union[float, int, ClientTimeout]
) -> Optional[float]:
    """
    Parse timeout

    :param timeout:
    """
    if isinstance(timeout, float):
        return timeout
    elif isinstance(timeout, int):
        return float(timeout)
    elif isinstance(timeout, ClientTimeout):
        return timeout.total
    else:
        raise ValueError("Timeout must be float, int or ClientTimeout. You "
                         f"passed {type(timeout)}")


class HistoryPollingMixin:
    """ Mixin, which provides polling """
    _requests: Any  # type: ignore

    def __init__(self, dispatcher: Any) -> None:
        self.dispatcher = dispatcher

    async def __parse_history_and_process_events(
            self,
            history: List[Transaction],
            last_payment_id: int
    ):
        """
        Processing events and send callbacks to handlers

        :param history: [list] transactions list
        :param last_payment_id: id of last payment in history
        """
        history_iterator = iter(history[::-1])

        while self.dispatcher.offset < last_payment_id:
            try:
                payment = next(history_iterator)
                await self.dispatcher.process_event(payment)
            except StopIteration:  # handle exhausted iterator
                break

            self.dispatcher.offset = payment.transaction_id
            self.dispatcher.offset_start_date = self.dispatcher.offset_end_date

    async def __pre_process(
            self,
            get_updates_from: Optional[datetime.datetime]
    ):
        """
        Pre process method, which set start date and end date of polling
        :param get_updates_from: date from which will be polling
        """
        try:
            current_time = datetime.datetime.now()
            assert isinstance(get_updates_from, datetime.datetime)
            assert (
                           current_time - get_updates_from
                   ).total_seconds() > 0
        except AssertionError as ex:
            raise ValueError(
                "Invalid value of get_updates_from, it must "
                "be instance of datetime and no more than  the current time"
            ) from ex

        self.dispatcher.offset_end_date = current_time

        if self.dispatcher.offset_start_date is None:
            self.dispatcher.offset_start_date = get_updates_from

    async def _get_history(self) -> List[Transaction]:
        """
        Get history by call 'transactions' method from QiwiWrapper.
        If history is empty or not all transactions not isinstance
         class Transaction - raise exception

        """
        history = await self.dispatcher.client.transactions(
            end_date=self.dispatcher.offset_end_date,
            start_date=self.dispatcher.offset_start_date
        )

        if not history or not all(
                isinstance(txn, Transaction) for txn in history):
            raise NoUpdatesToExecute()

        return history

    async def _pool_process(
            self,
            get_updates_from: Optional[datetime.datetime]
    ):
        """
        Method, which manage pool process

        :param get_updates_from: date from which will be polling
        """
        await self.__pre_process(get_updates_from)
        try:
            history = await self._get_history()
        except NoUpdatesToExecute:
            return

        last_payment = history[0]
        last_txn_id = last_payment.transaction_id

        if self.dispatcher.offset is None:
            first_payment = history[-1]
            self.dispatcher.offset = first_payment.transaction_id - 1

        await self.__parse_history_and_process_events(
            history=history,
            last_payment_id=last_txn_id
        )

    async def __start_polling(self, **kwargs):
        """
        Blocking method, which start polling process

        :param kwargs:
        """
        self.dispatcher._polling = True
        self.dispatcher.request_timeout = parse_timeout(kwargs.pop("timeout"))
        while self.dispatcher._polling:
            await self._pool_process(**kwargs)
            await asyncio.sleep(self.dispatcher.request_timeout)

    def __on_shutdown(self, loop: asyncio.AbstractEventLoop):
        """
        On shutdown we gracefully cancel all tasks, close event loop
        and call __aexit__ method to close aiohttp session
        """
        loop.run_until_complete(self.dispatcher.goodbye())
        api_helper.sync(self.__aexit__, None, None, None)
        api_helper.safe_cancel(loop)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Закрываем сессию и очищаем кэш при выходе"""
        if self._requests.session:
            await self._requests.session.close()
            self._requests.clear_cache()

    def start_polling(
            self,
            get_updates_from: datetime.datetime = datetime.datetime.now(),
            timeout: Union[float, int, ClientTimeout] = DEFAULT_TIMEOUT,
            on_startup: Optional[Callable] = None,
            on_shutdown: Optional[Callable] = None
    ):
        """
        Start long-polling mode

        :param get_updates_from: date from which will be polling,
         if it's None, polling will skip all updates
        :param timeout: timeout of polling in seconds, if the timeout is less,
         the API can throw an exception
        :param on_startup: function or coroutine,
         which will be executed on startup
        :param on_shutdown: function or coroutine,
         which will be executed on shutdown
         """
        self.dispatcher.logger.info("Start polling!")
        loop = api_helper.take_event_loop()
        _setup_callbacks(self.dispatcher, on_startup, on_shutdown)
        try:
            loop.run_until_complete(self.dispatcher.welcome())
            loop.create_task(self.__start_polling(
                get_updates_from=get_updates_from,
                timeout=timeout
            ))
            api_helper.run_forever_safe(loop=loop)
        except (SystemExit, KeyboardInterrupt):
            pass
        finally:
            self.__on_shutdown(loop)



import time
import typing

from glQiwiApi.core import BaseStorage
from glQiwiApi.core.constants import uncached
from glQiwiApi.types import MEMData
from glQiwiApi.types.basics import Cached, Attributes
from glQiwiApi.utils.exceptions import InvalidData


class Storage(BaseStorage):
    """
    Класс, позволяющий кэшировать результаты запросов

    """
    # Доступные критерии, по которым проходит валидацию кэш
    available = ('params', 'json', 'data', 'headers')

    __slots__ = ('data', '_cache_time', '_default_key')

    def __init__(
            self,
            cache_time: typing.Union[float, int],
            default_key: typing.Optional[str] = None
    ) -> None:
        """

        :param cache_time: Время кэширования в секундах
        :param default_key: дефолтный ключ, по которому будет доставаться кэш
        """
        if isinstance(cache_time, (int, float)):
            if cache_time > 60 or cache_time < 0:
                raise InvalidData(
                    "Время кэширования должно быть в пределах"
                    " от 0 до 60 секунд"
                )

        self.data: MEMData = MEMData({})
        self._cache_time = cache_time
        self.__initialize_default_key(default_key)

    def __initialize_default_key(self, key: typing.Optional[str]) -> None:
        """ Initialize default_key attribute """
        if not isinstance(key, str):
            self._default_key = "url"
        else:
            self._default_key = key
        self.data.update({
            self._default_key: {}
        })

    def get_current(self, key: typing.Any) -> typing.Any:
        """ Method to get element by key from data """
        try:
            obj = self.data.get(self._default_key).get(key)
            return obj if not self._is_expire(obj) else None
        except AttributeError:
            return None

    def clear(self, key: typing.Optional[str] = None,
              force: bool = False) -> typing.Any:
        """
        Method to delete element from the cache by key,
        or if force passed on its clear all data from the cache

        """
        if force:
            return self.data.clear()
        key: str
        return self.data[self._default_key].pop(key)

    def __setitem__(self, key, value) -> None:
        self.data[self._default_key][key] = value

    def __getitem__(self, item) -> typing.Any:
        try:
            return self.data[self._default_key][item]
        except KeyError:
            return None

    def _is_contain_uncached(self, value: typing.Optional[typing.Any]) -> bool:
        if self._cache_time < 0.1:
            return True

        for coincidence in uncached:
            try:
                if value.startswith(coincidence) or coincidence in value:
                    return True
            except AttributeError:
                return False
        return False

    def initialize_response_to_resolve(
            self,
            result: typing.Any,
            kwargs: dict,
            status_code: typing.Union[str, int]
    ) -> Cached:
        value = kwargs.get(self._default_key)
        if not self._is_contain_uncached(value):
            return Cached(
                kwargs=Attributes.format(kwargs, self.available),
                response_data=result,
                key=self._default_key,
                status_code=status_code,
                method=kwargs.get('method'),
                cache_to=value
            )
        if uncached[1] in value:
            self.clear(value, True)

    def update_data(self, obj_to_cache: typing.Any, key: typing.Any) -> None:
        """
        Метод, который добавляет результат запроса в кэш

        :param obj_to_cache: объект для кэширования
        :param key: ключ, по которому будет зарезервирован этот кэш

        """
        if isinstance(obj_to_cache, Cached):
            key = obj_to_cache.cache_to
        else:
            try:
                setattr(obj_to_cache, 'cached_in', time.monotonic())
                setattr(obj_to_cache, 'cache_to', key)
            except AttributeError:
                pass
        if not self._is_contain_uncached(obj_to_cache):
            self.data[self._default_key][key] = obj_to_cache

    @staticmethod
    def _check_get_request(cached: Cached, kwargs: dict) -> bool:
        """ Method to check cached get requests"""
        if cached.method == 'GET':
            if kwargs.get('headers') == cached.kwargs.headers:
                if kwargs.get('params') == cached.kwargs.params:
                    return True
        return False

    def _is_expire(self, cached: typing.Any) -> bool:
        """ Method to check live cache time, and if it expired return True """
        if time.monotonic() - cached.cached_in > self._cache_time:
            self.clear(cached.cache_to)
            return True
        return False

    def _validate_other(self, cached: Cached, kwargs: dict) -> bool:
        keys = (key for key in self.available if key != 'headers')
        for key in keys:
            if getattr(cached.kwargs, key) == kwargs.get(key, ''):
                return True
        return False

    def validate(self, kwargs: typing.Dict[str, typing.Any]) -> bool:
        """
        Метод, который по некоторым условиям
        проверяет актуальность кэша и в некоторых
        случая его чистит.

        :param kwargs:
        :return: boolean, прошел ли кэшированный запрос валидацию
        """
        # Если параметры и ссылка запроса совпадает
        cached = self.get_current(kwargs.get(self._default_key))
        if isinstance(cached, Cached):
            if self._is_expire(cached):
                return False

            # Проверяем запрос методом GET на кэш
            if self._check_get_request(cached, kwargs):
                return True

            elif self._validate_other(cached, kwargs):
                return True

        return False



from asyncio import as_completed, set_event_loop_policy
from itertools import repeat
from typing import (
    Dict,
    AsyncGenerator,
    NoReturn, Coroutine, Any
)
from typing import Optional, List, Union

import aiohttp
from aiohttp import (
    ClientTimeout,
    ClientProxyConnectionError,
    ServerDisconnectedError,
    ContentTypeError
)
from aiohttp.hdrs import USER_AGENT
from aiohttp.typedefs import LooseCookies

from glQiwiApi.core import AbstractParser
from glQiwiApi.core.constants import DEFAULT_TIMEOUT
from glQiwiApi.types import Response
from glQiwiApi.types.basics import Cached


class HttpXParser(AbstractParser):
    """
    Обвертка над aiohttp

    """

    _sleep_time = 2

    def __init__(self) -> None:
        self.base_headers = {
            'User-Agent': USER_AGENT,
            'Accept-Language': "ru-RU,ru;q=0.9,en-US;q=0.8,en;q=0.7"
        }
        self._timeout = ClientTimeout(
            total=5,
            connect=None,
            sock_connect=5,
            sock_read=None
        )

    async def _request(
            self,
            url: str,
            get_json: bool = False,
            method: str = 'POST',
            set_timeout: bool = True,
            cookies: Optional[LooseCookies] = None,
            json: Optional[dict] = None,
            data: Optional[Dict[str, Union[
                str, int, List[
                    Union[str, int]
                ]]]
            ] = None,
            headers: Optional[Dict[str, Union[str, int]]] = None,
            params: Optional[
                Dict[str, Union[str, int, List[
                    Union[str, int]
                ]]]
            ] = None) -> Response:
        """
        Метод для отправки запроса,
        может возвращать в Response ProxyError в качестве response_data,
        это означает, что вы имеете проблемы с подключением к прокси,
        возможно нужно добавить дополнительные post данные,
        если вы используете method = POST, или headers,
        если запрос GET


        :param url: ссылка, куда вы хотите отправить ваш запрос
        :param get_json: указывает на то, хотите ли вы получить ответ
         в формате json
        :param method: Тип запроса
        :param data: payload data
        :param cookies:
        :param headers:
         aiohttp.ClientSession initialization
        :return: Response instance
        """
        headers = self.get_headers(headers)
        # Create new session if old was closed
        self.create_session(
            timeout=self._timeout if set_timeout else DEFAULT_TIMEOUT
        )

        # sending query to some endpoint url
        try:
            response = await self.session.request(
                method=method,
                url=url,
                data=data,
                headers=headers,
                json=json if isinstance(json, dict) else None,
                cookies=cookies,
                params=params,
            )
        except (
                ClientProxyConnectionError,
                ServerDisconnectedError
        ):
            self.raise_exception(status_code='400_special_bad_proxy')
            return Response.bad_response()
        # Get content and return response
        try:
            data = await response.json(
                content_type="application/json"
            )
        except ContentTypeError:
            if get_json:
                return Response(status_code=response.status)
            data = await response.read()
        return Response(
            status_code=response.status,
            response_data=data,
            raw_headers=response.raw_headers,
            cookies=response.cookies,
            ok=response.ok,
            content_type=response.content_type,
            host=response.host,
            url=response.url.__str__()
        )

    def get_headers(self, headers: Optional[dict]) -> Optional[dict]:
        if isinstance(headers, dict):
            return headers
        return self.base_headers

    async def fetch(
            self,
            *,
            times: int = 1,
            **kwargs
    ) -> AsyncGenerator[Response, None]:
        """
        Basic usage: \n
        parser = HttpXParser() \n
        async for response in parser.fetch():
            print(response)

        :param times: int of quantity requests
        :param kwargs: HttpXParser._request kwargs
        :return:
        """
        coroutines = [self._request(**kwargs) for _ in repeat(None, times)]
        for future in as_completed(fs=coroutines):
            yield await future

    def fast(self) -> 'HttpXParser':
        """
        Method to fetching faster with using faster event loop(uvloop) \n
        USE IT ONLY ON LINUX SYSTEMS,
        on Windows or Mac its dont give performance!

        :return:
        """
        try:
            from uvloop import EventLoopPolicy  # type: ignore
            set_event_loop_policy(EventLoopPolicy())
        except ImportError:
            # Catching import error and forsake standard policy
            from asyncio import DefaultEventLoopPolicy as EventLoopPolicy
            set_event_loop_policy(EventLoopPolicy())
        return self



from .abstracts import (
    AbstractParser,
    BaseStorage,
    AioTestCase
)
from .aiohttp_custom_api import RequestManager
from .basic_requests_api import HttpXParser
from .core_mixins import BillMixin, ToolsMixin
from .storage import Storage

__all__ = (
    'HttpXParser',
    'Storage',
    'AbstractParser',
    'BaseStorage',
    'AioTestCase',
    'BillMixin',
    'ToolsMixin',
    'RequestManager'
)



import asyncio
import datetime
from typing import Dict

import pytest

from glQiwiApi import QiwiWrapper, InvalidData

pytestmark = pytest.mark.asyncio

CACHE_TIME = 5


@pytest.fixture(name='api')
async def api_fixture(
        credentials: Dict[str, str]
):
    """ Api fixture """
    _wrapper = QiwiWrapper(**credentials, cache_time=CACHE_TIME)
    yield _wrapper
    await _wrapper.close()


class TestCache:

    @pytest.mark.parametrize("cache_time", [-5, 70, 65])
    async def test_wrong_initialize_api(
            self,
            credentials: Dict[str, str],
            cache_time: int
    ):
        with pytest.raises(InvalidData):
            QiwiWrapper(**credentials, cache_time=cache_time)

    @pytest.mark.parametrize("payload", [
        {"rows_num": 50},
        {"rows_num": 50,
         "operation": "IN"},
        {
            "rows_num": 50,
            "operation": "IN",
            "start_date": datetime.datetime.now() - datetime.timedelta(
                days=50),
            "end_date": datetime.datetime.now()
        }
    ])
    async def test_query_caching(self, api: QiwiWrapper, payload: dict, capsys):
        async with api:
            first_response = await api.transactions(**payload)
            second_response = await asyncio.wait_for(
                api.transactions(
                    **payload
                ), 0.07  # We are waiting for task for 7 milliseconds,
                # because if the query has cached, it will just take
                # it from storage and API will take less time than query,
                # so there if query wont work we will see TimeoutError
            )

        assert first_response == second_response

    @pytest.mark.parametrize("payload1,payload2", [
        ({"rows_num": 50}, {"rows_num": 40}),
        ({"rows_num": 50,
          "operation": "IN"}, {"rows_num": 50, "operation": "OUT"}),
    ])
    async def test_uncached(
            self,
            api: QiwiWrapper,
            payload1: dict,
            payload2: dict
    ):
        async with api:
            first_response = await api.transactions(**payload1)
            with pytest.raises(asyncio.TimeoutError):
                await asyncio.wait_for(
                    api.transactions(**payload2), 0.05
                )
            second_uncached_response = await api.transactions(
                **payload2
            )
        assert first_response != second_uncached_response

    @pytest.mark.parametrize("payload", [
        {"rows_num": 50},
        {"rows_num": 50,
         "operation": "IN"},
        {
            "rows_num": 50,
            "operation": "IN",
            "start_date": datetime.datetime.now() - datetime.timedelta(
                days=50),
            "end_date": datetime.datetime.now()
        }
    ])
    async def test_response_cache_time(
            self,
            api: QiwiWrapper,
            payload: dict
    ):
        async with api:
            await api.transactions(**payload)
            await asyncio.sleep(CACHE_TIME + 1)
            with pytest.raises(asyncio.TimeoutError):
                await asyncio.wait_for(
                    api.transactions(
                        **payload
                    ), 0.07)



import asyncio
import inspect
import logging
import types
from datetime import datetime, timedelta
from typing import List, Tuple, Coroutine, Any, Union, Optional, NoReturn, \
    Callable

import aiohttp

from .config import EventHandlerFunctor, EventFilter, E
from .filter import Filter, transaction_webhook_filter, bill_webhook_filter


async def _inspect_and_execute_callback(client, callback: Callable):
    if inspect.iscoroutinefunction(callback):
        await callback(client)
    else:
        callback(client)


def _get_stream_handler() -> logging.StreamHandler:
    _log_format = f"%(asctime)s - [%(levelname)s] - %(message)s"
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.INFO)
    stream_handler.setFormatter(logging.Formatter(_log_format))
    return stream_handler


def _setup_logger() -> logging.Logger:
    logger = logging.getLogger(__name__)
    logger.setLevel(level=logging.INFO)
    if not logger.handlers:
        logger.addHandler(_get_stream_handler())
    return logger


class EventHandler:
    """
    Event handler, which executing and working with handlers

    """

    def __init__(self, functor: EventHandlerFunctor, *filter_: Filter) -> None:
        """

        :param functor:
        :param filter_:
        """
        self._fn = functor
        self._filters = filter_

    async def check_then_execute(self, event: E) -> Coroutine[Any, Any, Any]:
        """
        Check event, apply all filters and than pass on to handler

        :param event: handler
        """
        for filter_ in self._filters:
            if filter_.awaitable:
                if not await filter_.function(event):
                    break
            else:
                if not filter_.function(event):
                    break
        else:
            return await self._fn(event)


class Dispatcher:
    """
    Class, which managing all handlers

    """

    def __init__(
            self,
            loop: asyncio.AbstractEventLoop,
            wallet: Any
    ):
        if not isinstance(loop, asyncio.AbstractEventLoop):
            raise ValueError(
                f"Listener must have its event loop implemented with"
                f" {asyncio.AbstractEventLoop!r}"
            )

        self.loop = loop
        self.transaction_handlers: List[EventHandler] = []
        self.bill_handlers: List[EventHandler] = []
        self._logger: logging.Logger = _setup_logger()
        # Polling variables
        self._polling: bool = False
        self.offset: Optional[int] = None
        self.offset_start_date: Optional[
            Union[
                datetime,
                timedelta
            ]
        ] = None
        self.offset_end_date = None
        self.client: Any = wallet
        self.request_timeout: Optional[
            Union[float, int, aiohttp.ClientTimeout]
        ] = None
        self._on_startup: List[Callable] = []
        self._on_shutdown: List[Callable] = []

    def register_transaction_handler(
            self,
            event_handler: EventHandlerFunctor,
            *filters: EventFilter
    ):
        self.transaction_handlers.append(
            self.wrap_handler(
                event_handler, filters,
                default_filter=transaction_webhook_filter
            )
        )

    def register_bill_handler(self, event_handler: EventHandlerFunctor,
                              *filters: EventFilter):
        self.bill_handlers.append(
            self.wrap_handler(
                event_handler, filters,
                default_filter=bill_webhook_filter
            )
        )

    @property
    def handlers(self) -> List[EventHandler]:
        """ Return all registered handlers """
        return [*self.bill_handlers, *self.transaction_handlers]

    @property
    def logger(self) -> logging.Logger:
        return self._logger

    @logger.setter
    def logger(self, logger: logging.Logger):
        self._logger = logger

    def __setitem__(self, key: str, callback: Callable):
        if key not in ["on_shutdown", "on_startup"]:
            raise ValueError()

        if not isinstance(callback, types.FunctionType):
            raise ValueError("Invalid type of callback")

        if key == "on_shutdown":
            self._on_shutdown.append(callback)
        else:
            self._on_startup.append(callback)

    @staticmethod
    def wrap_handler(
            event_handler: EventHandlerFunctor,
            filters: Tuple[Union[EventFilter, Filter], ...],
            default_filter: Filter
    ) -> EventHandler:
        """
        Add new event handler.
        (!) Allows chain addition.
        :param event_handler: event handler, low order function
         which works with events
        :param filters: filter for low order function execution
        :param default_filter: default filter for handler
        :return: this handler manager
        """
        if filters:  # Initially filters are in tuple
            filters_list = list(filters)

            for index, filter_ in enumerate(filters):
                if not isinstance(filter_, Filter):
                    filters_list[index] = default_filter & Filter(filter_)

            filters = tuple(filters_list)
        else:
            filters = (default_filter,)

        return EventHandler(event_handler, *filters)

    def transaction_handler_wrapper(self, *filters: EventFilter) -> E:

        def decorator(callback: EventHandlerFunctor) -> EventHandlerFunctor:
            self.register_transaction_handler(callback, *filters)
            return callback

        return decorator

    def bill_handler_wrapper(self, *filters: EventFilter) -> E:

        def decorator(callback: EventHandlerFunctor) -> EventHandlerFunctor:
            self.register_bill_handler(callback, *filters)
            return callback

        return decorator

    async def process_event(self, event: E) -> None:
        """
        Feed handlers with event.
        :param event: any object that will be translated to handlers
        """

        for handler in self.handlers:
            await handler.check_then_execute(event)

    async def welcome(self) -> None:
        """"""
        for callback in self._on_startup:
            await _inspect_and_execute_callback(
                callback=callback,
                client=self.client
            )

    async def goodbye(self) -> None:
        self.logger.info("Stop polling!")
        for callback in self._on_shutdown:
            await _inspect_and_execute_callback(
                callback=callback,
                client=self.client
            )



import typing

from glQiwiApi import types
from glQiwiApi.core import RequestManager, ToolsMixin
from glQiwiApi.types.basics import DEFAULT_CACHE_TIME
from glQiwiApi.utils import basics as api_helper


class QiwiMaps(ToolsMixin):
    """
    API Карты терминалов QIWI позволяет установить местонахождение
    терминалов QIWI на территории РФ

    """

    def __init__(
            self,
            without_context: bool = False,
            cache_time: int = DEFAULT_CACHE_TIME
    ) -> None:
        self._requests = RequestManager(
            without_context=without_context,
            cache_time=cache_time
        )

    async def terminals(
            self,
            polygon: types.Polygon,
            zoom: int = None,
            pop_if_inactive_x_mins: int = 30,
            include_partners: bool = None,
            partners_ids: list = None,
            cache_terminals: bool = None,
            card_terminals: bool = None,
            identification_types: int = None,
            terminal_groups: list = None,
    ) -> typing.List[types.Terminal]:
        """
        Get map of terminals sent for passed polygon with additional params

        :param polygon: glQiwiApi.types.Polygon object
        :param zoom:
         https://tech.yandex.ru/maps/doc/staticapi/1.x/dg/concepts/map_scale-docpage/
        :param pop_if_inactive_x_mins: do not show if terminal
         was inactive for X minutes default 0.5 hours
        :param include_partners: result will include/exclude partner terminals
        :param partners_ids: Not fixed IDS array look at docs
        :param cache_terminals: result will include or exclude
         cache-acceptable terminals
        :param card_terminals: result will include or
         exclude card-acceptable terminals
        :param identification_types: `0` - not identified, `1` -
         partly identified, `2` - fully identified
        :param terminal_groups: look at QiwiMaps.partners
        :return: list of Terminal instances
        """
        params = self._requests.filter_dict(
            {
                **polygon.dict,
                "zoom": zoom,
                "activeWithinMinutes": pop_if_inactive_x_mins,
                "withRefillWallet": include_partners,
                "ttpIds": partners_ids,
                "cacheAllowed": cache_terminals,
                "cardAllowed": card_terminals,
                "identificationTypes": identification_types,
                "ttpGroups": terminal_groups,
            }
        )
        url = "http://edge.qiwi.com/locator/v3/nearest/clusters?parameters"
        async for response in self._requests.fast().fetch(
                url=url,
                method='GET',
                params=params
        ):
            return api_helper.multiply_objects_parse(
                lst_of_objects=response.response_data,
                model=types.Terminal
            )

    async def partners(self) -> typing.List[types.Partner]:
        """
        Get terminal partners for ttpGroups
        :return: list of TTPGroups
        """
        async for response in self._requests.fast().fetch(
                url='http://edge.qiwi.com/locator/v3/ttp-groups',
                method='GET',
                headers={"Content-type": "text/json"}
        ):
            return api_helper.multiply_objects_parse(
                lst_of_objects=response.response_data,
                model=types.Partner
            )



API_DATA = {
    "api_access_token": "2ef55757d94f77534a49251ee41c8d59",
    "secret_p2p":
        "eyJ2ZXJzaW9uIjoiUDJQIiwiZGF0YSI6eyJwYXlpbl9tZXJjaGFudF9zaXRlX3VpZCI6ImJuMXZmNy0wMCIsInVzZXJfaWQiOiIzODA5NjgzMTc0NTkiLCJzZWNyZXQiOiI4ZjI4NGVjYWQ0ZTE0Y2MwYzA5ZTRlOWNiNTJjM2Q3MzU2NGVjMWQxZDYyNWIwZDZhMTQ3NjIyZDEyZTJmNWFlIn19",
    "phone_number": "+380968317459"
}

YOO_MONEY_DATA = {
    "api_access_token": "4100116602400968.3CA5CF3053F6D047025F1BB937471BADD188C967D14C23F97F282FDC7697B574891F979175E072A4ABFAA5C91AE917C73D6781F6F3762058B45FAB1126C2CC6AC15D9306C6C767A84D9CF49C88DC9F7C12179C74CA75EA57727BCCE7564988A84E7529A19614E704F681337D97BE4A45A4E62E52EDDFE3C3A458BA16F6B6004F"
}

WRONG_API_DATA = {
    "api_access_token": dict(),
    "secret_p2p": 5454,
    "phone_number": 1234
}

TO_WALLET_DATA = {
    "to_number": "+380985272064",
    "trans_sum": 1,
    "comment": "unit_test"
}

NOTIFICATION_RAW_DATA = """{
"bill": {
    "siteId": "9hh4jb-00",
    "billId": "cc961e8d-d4d6-4f02-b737-2297e51fb48e",
    "amount": {
      "value": "1.00",
      "currency": "RUB"
    },
    "status": {
      "value": "PAID",
      "changedDateTime": "2021-01-18T15:25:18+03"
    },
    "customer": {
      "phone": "78710009999",
      "email": "test@tester.com",
      "account": "454678"
    },
    "customFields": {
      "paySourcesFilter": "qw",
      "themeCode": "Yvan-YKaSh",
      "yourParam1": "64728940",
      "yourParam2": "order 678"
    },
    "comment": "Text comment",
    "creationDateTime": "2021-01-18T15:24:53+03",
    "expirationDateTime": "2025-12-10T09:02:00+03"
  },
  "version": "1"
}"""

WEBHOOK_RAW_DATA = """{
    "hash": "50779a03d90c4fa60ac44dfd158dbceec0e9c57fa4cf4f5298450fdde1868945",
    "hookId": "f57f95e2-149f-4278-b2cb-4114bc319727",
    "messageId": "f9a197a8-26b6-4d42-aac4-d86b789c373c",
    "payment": {"account": "thedandod",
                 "comment": "",
                 "date": "2018-05-18T16:05:15+03:00",
                 "errorCode": "0",
                 "personId": 79254914194,
                 "provider": 25549,
                 "signFields": "sum.currency,sum.amount,type,account,txnId",
                 "status": "WAITING",
                 "sum": {"amount": 1.73, "currency": 643},
                 "total": {"amount": 1.73, "currency": 643},
                 "txnId": "13117338074",
                 "type": "OUT"},
    "test": false,
    "version": "1.0.0"
    }"""

TXN_RAW_DATA = """{
    "txnId":9309,
    "personId":79112223344,
    "date":"2017-01-21T11:41:07+03:00",
    "errorCode":0,
    "error":null,
    "status":"SUCCESS",
    "type":"OUT",
    "statusText":"Успешно",
    "trmTxnId":"1489826461807",
    "account":"0003***",
    "sum":{
        "amount":70,
        "currency":"RUB"
        },
    "commission":{
        "amount":0,
        "currency":"RUB"
        },
    "total":{
        "amount":70,
        "currency":"RUB"
        },
    "provider":{
      "id":5,
      "shortName":"test_provider"
    },
    "source": {},
    "comment":null,
    "currencyRate":1,
    "extras":null,
    "chequeReady":true,
    "bankDocumentAvailable":false,
    "bankDocumentReady":false,
    "repeatPaymentEnabled":false,
    "favoritePaymentEnabled": true,
    "regularPaymentEnabled": true
   }
"""

TEMP_DIRECTORY_NAME: str = 'temp'

RECEIPT_FILE_NAME: str = 'receipt'



from typing import Dict

import pytest
import timeout_decorator

from glQiwiApi import QiwiWrapper
from glQiwiApi import types


@pytest.fixture(name='credentials')
def credentials_fixture():
    """ credentials fixture """
    from ..types.dataset import API_DATA
    yield API_DATA


@pytest.mark.asyncio
@pytest.fixture(name='api')
async def api_fixture(credentials: Dict[str, str]):
    """ Api fixture """
    _wrapper = QiwiWrapper(**credentials)
    yield _wrapper
    await _wrapper.close()


async def _on_startup_callback(api: QiwiWrapper):
    from ..types.dataset import TO_WALLET_DATA
    await api.to_wallet(**TO_WALLET_DATA)


class TestPolling:

    @timeout_decorator.timeout(5)
    def _start_polling(self, api: QiwiWrapper):
        self._handled = False

        # Also, without decorators, you can do like this
        # api.dispatcher.register_transaction_handler(my_handler)
        @api.transaction_handler()
        async def my_handler(event: types.Transaction):
            self._handled = True
            assert isinstance(event, types.Transaction)

        api.start_polling(
            timeout=10,
            on_startup=_on_startup_callback
        )

    def test_polling(self, api: QiwiWrapper):
        try:
            self._start_polling(api)
        except timeout_decorator.TimeoutError:
            assert self._handled is True



from typing import Union

from pydantic import BaseModel, Field, validator

from glQiwiApi.types.qiwi_types.currency_parsed import CurrencyModel
from glQiwiApi.utils.currency_util import Currency


class CrossRate(BaseModel):
    """Курс валюты"""
    rate_from: Union[str, CurrencyModel] = Field(..., alias="from")
    rate_to: Union[str, CurrencyModel] = Field(..., alias="to")
    rate: float

    @validator("rate_from", "rate_to", pre=True)
    def humanize_rates(cls, v):
        if not isinstance(v, str):
            return v
        cur = Currency.get(v)
        if not cur:
            return v
        return cur


class PaymentMethod(BaseModel):
    payment_type: str
    account_id: str


class FreePaymentDetailsFields(BaseModel):
    """ Набор реквизитов платежа"""
    name: str
    """Наименование банка получателя"""

    extra_to_bik: str
    """БИК банка получателя"""

    to_bik: str
    """	БИК банка получателя"""

    city: str
    """Город местонахождения получателя"""

    info: str = 'Коммерческие организации'
    """Константное значение"""

    is_commercial: str = "1"
    """Служебная информация"""

    to_name: str
    """Наименование организации"""

    to_inn: str
    """ИНН организации"""

    to_kpp: str
    """	КПП организации"""

    nds: str
    """
    Признак уплаты НДС.
    Если вы оплачиваете квитанцию и в ней не указан НДС,
    то строка НДС не облагается. В ином случае, строка В т.ч. НДС
    """

    goal: str
    """Назначение платежа"""

    urgent: str = "0"
    """
    Признак срочного платежа (0 - нет, 1 - да).
    Срочный платеж выполняется от 10 минут.
    Возможен по будням с 9:00 до 20:30 по московскому времени. 
    Стоимость услуги — 25 рублей.
    """

    account: str
    """Номер счета получателя"""

    from_name: str
    """Имя плательщика"""

    from_name_p: str
    """Отчество плательщика"""

    from_name_f: str
    """	Фамилия плательщика"""

    requestProtocol: str = "qw1"
    """Служебная информация, константа"""

    toServiceId: str = "1717"
    """Служебная информация, константа"""


__all__ = ('CrossRate', 'FreePaymentDetailsFields', 'PaymentMethod')
